from django import forms
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.http import Http404
from django.utils import timezone
from django.utils.translation import npgettext, pgettext_lazy

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role
from ...acl.objectacl import add_acl_to_obj
from ...admin.forms import YesNoSwitch
from ...categories.models import Category, CategoryRole
from ...categories.permissions import get_categories_roles
from ..models import Post, Thread

__all__ = [
    "allow_see_thread",
    "can_see_thread",
    "allow_start_thread",
    "can_start_thread",
    "allow_reply_thread",
    "can_reply_thread",
    "allow_edit_thread",
    "can_edit_thread",
    "allow_pin_thread",
    "can_pin_thread",
    "allow_unhide_thread",
    "can_unhide_thread",
    "allow_hide_thread",
    "can_hide_thread",
    "allow_delete_thread",
    "can_delete_thread",
    "allow_move_thread",
    "can_move_thread",
    "allow_merge_thread",
    "can_merge_thread",
    "allow_approve_thread",
    "can_approve_thread",
    "allow_see_post",
    "can_see_post",
    "allow_edit_post",
    "can_edit_post",
    "allow_unhide_post",
    "can_unhide_post",
    "allow_hide_post",
    "can_hide_post",
    "allow_delete_post",
    "can_delete_post",
    "allow_protect_post",
    "can_protect_post",
    "allow_approve_post",
    "can_approve_post",
    "allow_move_post",
    "can_move_post",
    "allow_merge_post",
    "can_merge_post",
    "allow_unhide_event",
    "can_unhide_event",
    "allow_split_post",
    "can_split_post",
    "allow_hide_event",
    "can_hide_event",
    "allow_delete_event",
    "can_delete_event",
    "exclude_invisible_threads",
    "exclude_invisible_posts",
]


class RolePermissionsForm(forms.Form):
    legend = pgettext_lazy("threads permission", "Threads")

    can_see_unapproved_content_lists = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can see unapproved content list"),
        help_text=pgettext_lazy(
            "threads permission",
            'Allows access to "unapproved" tab on threads lists for easy listing of threads that are unapproved or contain unapproved posts. Despite the tab being available on all threads lists, it will only display threads belonging to categories in which the user has permission to approve content.',
        ),
    )
    can_see_reported_content_lists = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can see reported content list"),
        help_text=pgettext_lazy(
            "threads permission",
            'Allows access to "reported" tab on threads lists for easy listing of threads that contain reported posts. Despite the tab being available on all categories threads lists, it will only display threads belonging to categories in which the user has permission to see posts reports.',
        ),
    )
    can_omit_flood_protection = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can omit flood protection"),
        help_text=pgettext_lazy(
            "threads permission",
            "Allows posting more frequently than flood protection would.",
        ),
    )


class CategoryPermissionsForm(forms.Form):
    legend = pgettext_lazy("threads permission", "Threads")

    can_see_all_threads = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can see threads"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads see permission choice", "Started threads")),
            (1, pgettext_lazy("threads see permission choice", "All threads")),
        ],
    )

    can_start_threads = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can start threads")
    )
    can_reply_threads = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can reply to threads")
    )

    can_edit_threads = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can edit threads"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads edit permission choice", "No")),
            (1, pgettext_lazy("threads edit permission choice", "Own threads")),
            (2, pgettext_lazy("threads edit permission choice", "All threads")),
        ],
    )
    can_hide_own_threads = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can hide own threads"),
        help_text=pgettext_lazy(
            "threads permission",
            "Only threads started within time limit and with no replies can be hidden.",
        ),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads hide own permission choice", "No")),
            (1, pgettext_lazy("threads hide own permission choice", "Hide threads")),
            (2, pgettext_lazy("threads hide own permission choice", "Delete threads")),
        ],
    )
    thread_edit_time = forms.IntegerField(
        label=pgettext_lazy(
            "threads permission", "Time limit for editing own threads, in minutes"
        ),
        help_text=pgettext_lazy(
            "threads permission", "Enter 0 to don't limit time for editing own threads."
        ),
        initial=0,
        min_value=0,
    )
    can_hide_threads = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can hide all threads"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads hide permission choice", "No")),
            (1, pgettext_lazy("threads hide permission choice", "Hide threads")),
            (2, pgettext_lazy("threads hide permission choice", "Delete threads")),
        ],
    )

    can_pin_threads = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can pin threads"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads pin permission choice", "No")),
            (1, pgettext_lazy("threads pin permission choice", "In category")),
            (2, pgettext_lazy("threads pin permission choice", "Globally")),
        ],
    )
    can_close_threads = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can close threads")
    )
    can_move_threads = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can move threads")
    )
    can_merge_threads = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can merge threads")
    )

    can_edit_posts = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can edit posts"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads posts edit permission choice", "No")),
            (1, pgettext_lazy("threads posts edit permission choice", "Own posts")),
            (2, pgettext_lazy("threads posts edit permission choice", "All posts")),
        ],
    )
    can_hide_own_posts = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can hide own posts"),
        help_text=pgettext_lazy(
            "threads permission",
            "Only last posts to thread made within edit time limit can be hidden.",
        ),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads own posts hide permission choice", "No")),
            (
                1,
                pgettext_lazy("threads own posts hide permission choice", "Hide posts"),
            ),
            (
                2,
                pgettext_lazy(
                    "threads own posts hide permission choice", "Delete posts"
                ),
            ),
        ],
    )
    post_edit_time = forms.IntegerField(
        label=pgettext_lazy(
            "threads permission", "Time limit for editing own post, in minutes"
        ),
        help_text=pgettext_lazy(
            "threads permission", "Enter 0 to don't limit time for editing own posts."
        ),
        initial=0,
        min_value=0,
    )
    can_hide_posts = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can hide all posts"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads hide posts permission choice", "No")),
            (1, pgettext_lazy("threads hide posts permission choice", "Hide posts")),
            (2, pgettext_lazy("threads hide posts permission choice", "Delete posts")),
        ],
    )

    can_see_posts_likes = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can see posts likes"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads posts like see permission choice", "No")),
            (
                1,
                pgettext_lazy(
                    "threads posts like see permission choice", "Number only"
                ),
            ),
            (
                2,
                pgettext_lazy(
                    "threads posts like see permission choice",
                    "Number and list of likers",
                ),
            ),
        ],
    )
    can_like_posts = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can like posts"),
        help_text=pgettext_lazy(
            "threads permission",
            "Only users with this permission to see likes can like posts.",
        ),
    )

    can_protect_posts = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can protect posts"),
        help_text=pgettext_lazy(
            "threads permission",
            "Only users with this permission can edit protected posts.",
        ),
    )
    can_move_posts = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can move posts"),
        help_text=pgettext_lazy(
            "threads permission", "Will be able to move posts to other threads."
        ),
    )
    can_merge_posts = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can merge posts")
    )
    can_approve_content = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can approve content"),
        help_text=pgettext_lazy(
            "threads permission", "Will be able to see and approve unapproved content."
        ),
    )
    can_report_content = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can report posts")
    )
    can_see_reports = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Can see reports")
    )

    can_hide_events = forms.TypedChoiceField(
        label=pgettext_lazy("threads permission", "Can hide events"),
        coerce=int,
        initial=0,
        choices=[
            (0, pgettext_lazy("threads events hide permission choice", "No")),
            (1, pgettext_lazy("threads events hide permission choice", "Hide events")),
            (
                2,
                pgettext_lazy("threads events hide permission choice", "Delete events"),
            ),
        ],
    )

    require_threads_approval = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Require threads approval")
    )
    require_replies_approval = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Require replies approval")
    )
    require_edits_approval = YesNoSwitch(
        label=pgettext_lazy("threads permission", "Require edits approval")
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return RolePermissionsForm
    if isinstance(role, CategoryRole):
        return CategoryPermissionsForm


def build_acl(acl, roles, key_name):
    acl.update(
        {
            "can_see_unapproved_content_lists": False,
            "can_see_reported_content_lists": False,
            "can_omit_flood_protection": False,
            "can_approve_content": [],
            "can_see_reports": [],
        }
    )

    acl = algebra.sum_acls(
        acl,
        roles=roles,
        key=key_name,
        can_see_unapproved_content_lists=algebra.greater,
        can_see_reported_content_lists=algebra.greater,
        can_omit_flood_protection=algebra.greater,
    )

    categories_roles = get_categories_roles(roles)
    categories = list(Category.objects.all_categories(include_root=True))

    for category in categories:
        category_acl = acl["categories"].get(category.pk, {"can_browse": 0})
        if category_acl["can_browse"]:
            category_acl = acl["categories"][category.pk] = build_category_acl(
                category_acl, category, categories_roles, key_name
            )

            if category_acl.get("can_approve_content"):
                acl["can_approve_content"].append(category.pk)
            if category_acl.get("can_see_reports"):
                acl["can_see_reports"].append(category.pk)

    return acl


def build_category_acl(acl, category, categories_roles, key_name):
    category_roles = categories_roles.get(category.pk, [])

    final_acl = {
        "can_see_all_threads": 0,
        "can_start_threads": 0,
        "can_reply_threads": 0,
        "can_edit_threads": 0,
        "can_edit_posts": 0,
        "can_hide_own_threads": 0,
        "can_hide_own_posts": 0,
        "thread_edit_time": 0,
        "post_edit_time": 0,
        "can_hide_threads": 0,
        "can_hide_posts": 0,
        "can_protect_posts": 0,
        "can_move_posts": 0,
        "can_merge_posts": 0,
        "can_pin_threads": 0,
        "can_close_threads": 0,
        "can_move_threads": 0,
        "can_merge_threads": 0,
        "can_report_content": 0,
        "can_see_reports": 0,
        "can_see_posts_likes": 0,
        "can_like_posts": 0,
        "can_approve_content": 0,
        "require_threads_approval": 0,
        "require_replies_approval": 0,
        "require_edits_approval": 0,
        "can_hide_events": 0,
    }
    final_acl.update(acl)

    algebra.sum_acls(
        final_acl,
        roles=category_roles,
        key=key_name,
        can_see_all_threads=algebra.greater,
        can_start_threads=algebra.greater,
        can_reply_threads=algebra.greater,
        can_edit_threads=algebra.greater,
        can_edit_posts=algebra.greater,
        can_hide_threads=algebra.greater,
        can_hide_posts=algebra.greater,
        can_hide_own_threads=algebra.greater,
        can_hide_own_posts=algebra.greater,
        thread_edit_time=algebra.greater_or_zero,
        post_edit_time=algebra.greater_or_zero,
        can_protect_posts=algebra.greater,
        can_move_posts=algebra.greater,
        can_merge_posts=algebra.greater,
        can_pin_threads=algebra.greater,
        can_close_threads=algebra.greater,
        can_move_threads=algebra.greater,
        can_merge_threads=algebra.greater,
        can_report_content=algebra.greater,
        can_see_reports=algebra.greater,
        can_see_posts_likes=algebra.greater,
        can_like_posts=algebra.greater,
        can_approve_content=algebra.greater,
        require_threads_approval=algebra.greater,
        require_replies_approval=algebra.greater,
        require_edits_approval=algebra.greater,
        can_hide_events=algebra.greater,
    )

    return final_acl


def add_acl_to_category(user_acl, category):
    category_acl = user_acl["categories"].get(category.pk, {})

    category.acl.update(
        {
            "can_see_all_threads": 0,
            "can_see_own_threads": 0,
            "can_start_threads": 0,
            "can_reply_threads": 0,
            "can_edit_threads": 0,
            "can_edit_posts": 0,
            "can_hide_own_threads": 0,
            "can_hide_own_posts": 0,
            "thread_edit_time": 0,
            "post_edit_time": 0,
            "can_hide_threads": 0,
            "can_hide_posts": 0,
            "can_protect_posts": 0,
            "can_move_posts": 0,
            "can_merge_posts": 0,
            "can_pin_threads": 0,
            "can_close_threads": 0,
            "can_move_threads": 0,
            "can_merge_threads": 0,
            "can_report_content": 0,
            "can_see_reports": 0,
            "can_see_posts_likes": 0,
            "can_like_posts": 0,
            "can_approve_content": 0,
            "require_threads_approval": category.require_threads_approval,
            "require_replies_approval": category.require_replies_approval,
            "require_edits_approval": category.require_edits_approval,
            "can_hide_events": 0,
        }
    )

    algebra.sum_acls(
        category.acl,
        acls=[category_acl],
        can_see_all_threads=algebra.greater,
        can_see_posts_likes=algebra.greater,
    )

    if user_acl["is_authenticated"]:
        algebra.sum_acls(
            category.acl,
            acls=[category_acl],
            can_start_threads=algebra.greater,
            can_reply_threads=algebra.greater,
            can_edit_threads=algebra.greater,
            can_edit_posts=algebra.greater,
            can_hide_threads=algebra.greater,
            can_hide_posts=algebra.greater,
            can_hide_own_threads=algebra.greater,
            can_hide_own_posts=algebra.greater,
            thread_edit_time=algebra.greater_or_zero,
            post_edit_time=algebra.greater_or_zero,
            can_protect_posts=algebra.greater,
            can_move_posts=algebra.greater,
            can_merge_posts=algebra.greater,
            can_pin_threads=algebra.greater,
            can_close_threads=algebra.greater,
            can_move_threads=algebra.greater,
            can_merge_threads=algebra.greater,
            can_report_content=algebra.greater,
            can_see_reports=algebra.greater,
            can_like_posts=algebra.greater,
            can_approve_content=algebra.greater,
            require_threads_approval=algebra.greater,
            require_replies_approval=algebra.greater,
            require_edits_approval=algebra.greater,
            can_hide_events=algebra.greater,
        )

    if user_acl["can_approve_content"]:
        category.acl.update(
            {
                "require_threads_approval": 0,
                "require_replies_approval": 0,
                "require_edits_approval": 0,
            }
        )

    category.acl["can_see_own_threads"] = not category.acl["can_see_all_threads"]


def add_acl_to_thread(user_acl, thread):
    category_acl = user_acl["categories"].get(thread.category_id, {})

    thread.acl.update(
        {
            "can_reply": can_reply_thread(user_acl, thread),
            "can_edit": can_edit_thread(user_acl, thread),
            "can_pin": can_pin_thread(user_acl, thread),
            "can_pin_globally": False,
            "can_hide": can_hide_thread(user_acl, thread),
            "can_unhide": can_unhide_thread(user_acl, thread),
            "can_delete": can_delete_thread(user_acl, thread),
            "can_close": category_acl.get("can_close_threads", False),
            "can_move": can_move_thread(user_acl, thread),
            "can_merge": can_merge_thread(user_acl, thread),
            "can_move_posts": category_acl.get("can_move_posts", False),
            "can_merge_posts": category_acl.get("can_merge_posts", False),
            "can_approve": can_approve_thread(user_acl, thread),
            "can_see_reports": category_acl.get("can_see_reports", False),
        }
    )

    if thread.acl["can_pin"] and category_acl.get("can_pin_threads") == 2:
        thread.acl["can_pin_globally"] = True


def add_acl_to_post(user_acl, post):
    if post.is_event:
        add_acl_to_event(user_acl, post)
    else:
        add_acl_to_reply(user_acl, post)


def add_acl_to_event(user_acl, event):
    can_hide_events = 0

    if user_acl["is_authenticated"]:
        category_acl = user_acl["categories"].get(
            event.category_id, {"can_hide_events": 0}
        )

        can_hide_events = category_acl["can_hide_events"]

    event.acl.update(
        {
            "can_see_hidden": can_hide_events > 0,
            "can_hide": can_hide_event(user_acl, event),
            "can_delete": can_delete_event(user_acl, event),
        }
    )


def add_acl_to_reply(user_acl, post):
    category_acl = user_acl["categories"].get(post.category_id, {})

    post.acl.update(
        {
            "can_reply": can_reply_thread(user_acl, post.thread),
            "can_edit": can_edit_post(user_acl, post),
            "can_see_hidden": post.is_first_post or category_acl.get("can_hide_posts"),
            "can_unhide": can_unhide_post(user_acl, post),
            "can_hide": can_hide_post(user_acl, post),
            "can_delete": can_delete_post(user_acl, post),
            "can_protect": can_protect_post(user_acl, post),
            "can_approve": can_approve_post(user_acl, post),
            "can_move": can_move_post(user_acl, post),
            "can_merge": can_merge_post(user_acl, post),
            "can_report": category_acl.get("can_report_content", False),
            "can_see_reports": category_acl.get("can_see_reports", False),
            "can_see_likes": category_acl.get("can_see_posts_likes", 0),
            "can_like": False,
            "can_see_protected": False,
        }
    )

    if not post.acl["can_see_hidden"]:
        post.acl["can_see_hidden"] = post.id == post.thread.first_post_id
    if user_acl["is_authenticated"] and post.acl["can_see_likes"]:
        post.acl["can_like"] = category_acl.get("can_like_posts", False)
    if user_acl["is_authenticated"]:
        post.acl["can_see_protected"] = (
            post.acl["can_protect"] or user_acl["user_id"] == post.poster_id
        )


def register_with(registry):
    registry.acl_annotator(Category, add_acl_to_category)
    registry.acl_annotator(Thread, add_acl_to_thread)
    registry.acl_annotator(Post, add_acl_to_post)


def allow_see_thread(user_acl, target):
    category_acl = user_acl["categories"].get(
        target.category_id, {"can_see": False, "can_browse": False}
    )

    if not (category_acl["can_see"] and category_acl["can_browse"]):
        raise Http404()

    if target.is_hidden and (
        user_acl["is_anonymous"] or not category_acl["can_hide_threads"]
    ):
        raise Http404()

    if user_acl["is_anonymous"] or user_acl["user_id"] != target.starter_id:
        if not category_acl["can_see_all_threads"]:
            raise Http404()

        if target.is_unapproved and not category_acl["can_approve_content"]:
            raise Http404()


can_see_thread = return_boolean(allow_see_thread)


def allow_start_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to start threads.")
        )

    category_acl = user_acl["categories"].get(target.pk, {"can_start_threads": False})

    if not category_acl["can_start_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "You don't have permission to start new threads in this category.",
            )
        )

    if target.is_closed and not category_acl["can_close_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "This category is closed. You can't start new threads in it.",
            )
        )


can_start_thread = return_boolean(allow_start_thread)


def allow_reply_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to reply threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_reply_threads": False}
    )

    if not category_acl["can_reply_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't reply to threads in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't reply to threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't reply to closed threads in this category.",
                )
            )


can_reply_thread = return_boolean(allow_reply_thread)


def allow_edit_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to edit threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_edit_threads": False}
    )

    if not category_acl["can_edit_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't edit threads in this category."
            )
        )

    if category_acl["can_edit_threads"] == 1:
        if user_acl["user_id"] != target.starter_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't edit other users threads in this category.",
                )
            )

        if not has_time_to_edit_thread(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't edit threads that are older than %(minutes)s minute.",
                "You can't edit threads that are older than %(minutes)s minutes.",
                category_acl["thread_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["thread_edit_time"]}
            )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't edit threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't edit it."
                )
            )


can_edit_thread = return_boolean(allow_edit_thread)


def allow_pin_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You have to sign in to change threads weights."
            )
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_pin_threads": 0}
    )

    if not category_acl["can_pin_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "You can't change threads weights in this category.",
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't change threads weights in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't change its weight.",
                )
            )


can_pin_thread = return_boolean(allow_pin_thread)


def allow_unhide_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to hide threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_close_threads": False}
    )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't reveal threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't reveal it."
                )
            )


can_unhide_thread = return_boolean(allow_unhide_thread)


def allow_hide_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to hide threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_threads": 0, "can_hide_own_threads": 0}
    )

    if (
        not category_acl["can_hide_threads"]
        and not category_acl["can_hide_own_threads"]
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't hide threads in this category."
            )
        )

    if not category_acl["can_hide_threads"] and category_acl["can_hide_own_threads"]:
        if user_acl["user_id"] != target.starter_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't hide other users theads in this category.",
                )
            )

        if not has_time_to_edit_thread(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't hide threads that are older than %(minutes)s minute.",
                "You can't hide threads that are older than %(minutes)s minutes.",
                category_acl["thread_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["thread_edit_time"]}
            )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't hide threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't hide it."
                )
            )


can_hide_thread = return_boolean(allow_hide_thread)


def allow_delete_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You have to sign in to delete threads."
            )
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_threads": 0, "can_hide_own_threads": 0}
    )

    if (
        category_acl["can_hide_threads"] != 2
        and category_acl["can_hide_own_threads"] != 2
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't delete threads in this category."
            )
        )

    if (
        category_acl["can_hide_threads"] != 2
        and category_acl["can_hide_own_threads"] == 2
    ):
        if user_acl["user_id"] != target.starter_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't delete other users theads in this category.",
                )
            )

        if not has_time_to_edit_thread(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't delete threads that are older than %(minutes)s minute.",
                "You can't delete threads that are older than %(minutes)s minutes.",
                category_acl["thread_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["thread_edit_time"]}
            )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't delete threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't delete it."
                )
            )


can_delete_thread = return_boolean(allow_delete_thread)


def allow_move_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to move threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_move_threads": 0}
    )

    if not category_acl["can_move_threads"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't move threads in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't move it's threads.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't move it."
                )
            )


can_move_thread = return_boolean(allow_move_thread)


def allow_merge_thread(user_acl, target, otherthread=False):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to merge threads.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_merge_threads": 0}
    )

    if not category_acl["can_merge_threads"]:
        if otherthread:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "Other thread can't be merged with."
                )
            )
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't merge threads in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            if otherthread:
                raise PermissionDenied(
                    pgettext_lazy(
                        "threads permission",
                        "Other thread's category is closed. You can't merge with it.",
                    )
                )
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't merge it's threads.",
                )
            )
        if target.is_closed:
            if otherthread:
                raise PermissionDenied(
                    pgettext_lazy(
                        "threads permission",
                        "Other thread is closed and can't be merged with.",
                    )
                )
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't merge it with other threads.",
                )
            )


can_merge_thread = return_boolean(allow_merge_thread)


def allow_approve_thread(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You have to sign in to approve threads."
            )
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_approve_content": 0}
    )

    if not category_acl["can_approve_content"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't approve threads in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't approve threads in it.",
                )
            )
        if target.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This thread is closed. You can't approve it."
                )
            )


can_approve_thread = return_boolean(allow_approve_thread)


def allow_see_post(user_acl, target):
    category_acl = user_acl["categories"].get(
        target.category_id, {"can_approve_content": False, "can_hide_events": False}
    )

    if not target.is_event and target.is_unapproved:
        if user_acl["is_anonymous"]:
            raise Http404()

        if (
            not category_acl["can_approve_content"]
            and user_acl["user_id"] != target.poster_id
        ):
            raise Http404()

    if target.is_event and target.is_hidden and not category_acl["can_hide_events"]:
        raise Http404()


can_see_post = return_boolean(allow_see_post)


def allow_edit_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to edit posts.")
        )

    if target.is_event:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "Events can't be edited.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_edit_posts": False}
    )

    if not category_acl["can_edit_posts"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't edit posts in this category."
            )
        )

    if (
        target.is_hidden
        and not target.is_first_post
        and not category_acl["can_hide_posts"]
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "This post is hidden, you can't edit it."
            )
        )

    if category_acl["can_edit_posts"] == 1:
        if target.poster_id != user_acl["user_id"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't edit other users posts in this category.",
                )
            )

        if target.is_protected and not category_acl["can_protect_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This post is protected. You can't edit it."
                )
            )

        if not has_time_to_edit_post(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't edit posts that are older than %(minutes)s minute.",
                "You can't edit posts that are older than %(minutes)s minutes.",
                category_acl["post_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["post_edit_time"]}
            )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't edit posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't edit posts in it.",
                )
            )


can_edit_post = return_boolean(allow_edit_post)


def allow_unhide_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to reveal posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_posts": 0, "can_hide_own_posts": 0}
    )

    if not category_acl["can_hide_posts"]:
        if not category_acl["can_hide_own_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "You can't reveal posts in this category."
                )
            )

        if user_acl["user_id"] != target.poster_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't reveal other users posts in this category.",
                )
            )

        if target.is_protected and not category_acl["can_protect_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This post is protected. You can't reveal it."
                )
            )

        if not has_time_to_edit_post(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't reveal posts that are older than %(minutes)s minute.",
                "You can't reveal posts that are older than %(minutes)s minutes.",
                category_acl["post_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["post_edit_time"]}
            )

    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "Thread's first post can only be revealed using the reveal thread option.",
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't reveal posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't reveal posts in it.",
                )
            )


can_unhide_post = return_boolean(allow_unhide_post)


def allow_hide_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to hide posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_posts": 0, "can_hide_own_posts": 0}
    )

    if not category_acl["can_hide_posts"]:
        if not category_acl["can_hide_own_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "You can't hide posts in this category."
                )
            )

        if user_acl["user_id"] != target.poster_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't hide other users posts in this category.",
                )
            )

        if target.is_protected and not category_acl["can_protect_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This post is protected. You can't hide it."
                )
            )

        if not has_time_to_edit_post(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't hide posts that are older than %(minutes)s minute.",
                "You can't hide posts that are older than %(minutes)s minutes.",
                category_acl["post_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["post_edit_time"]}
            )

    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "Thread's first post can only be hidden using the hide thread option.",
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't hide posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't hide posts in it.",
                )
            )


can_hide_post = return_boolean(allow_hide_post)


def allow_delete_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to delete posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_posts": 0, "can_hide_own_posts": 0}
    )

    if category_acl["can_hide_posts"] != 2:
        if category_acl["can_hide_own_posts"] != 2:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "You can't delete posts in this category."
                )
            )

        if user_acl["user_id"] != target.poster_id:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "You can't delete other users posts in this category.",
                )
            )

        if target.is_protected and not category_acl["can_protect_posts"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission", "This post is protected. You can't delete it."
                )
            )

        if not has_time_to_edit_post(user_acl, target):
            message = npgettext(
                "threads permission",
                "You can't delete posts that are older than %(minutes)s minute.",
                "You can't delete posts that are older than %(minutes)s minutes.",
                category_acl["post_edit_time"],
            )
            raise PermissionDenied(
                message % {"minutes": category_acl["post_edit_time"]}
            )

    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "Thread's first post can only be deleted together with thread.",
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't delete posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't delete posts in it.",
                )
            )


can_delete_post = return_boolean(allow_delete_post)


def allow_protect_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to protect posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_protect_posts": False}
    )

    if not category_acl["can_protect_posts"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't protect posts in this category."
            )
        )
    if not can_edit_post(user_acl, target):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't protect posts you can't edit."
            )
        )


can_protect_post = return_boolean(allow_protect_post)


def allow_approve_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to approve posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_approve_content": False}
    )

    if not category_acl["can_approve_content"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't approve posts in this category."
            )
        )
    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "Thread's first post can only be approved together with thread.",
            )
        )
    if (
        not target.is_first_post
        and not category_acl["can_hide_posts"]
        and target.is_hidden
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "You can't approve posts the content you can't see.",
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't approve posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't approve posts in it.",
                )
            )


can_approve_post = return_boolean(allow_approve_post)


def allow_move_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to move posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_move_posts": False}
    )

    if not category_acl["can_move_posts"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't move posts in this category."
            )
        )
    if target.is_event:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "Events can't be moved.")
        )
    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission",
                "Thread's first post can only be moved together with thread.",
            )
        )
    if not category_acl["can_hide_posts"] and target.is_hidden:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't move posts the content you can't see."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't move posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't move posts in it.",
                )
            )


can_move_post = return_boolean(allow_move_post)


def allow_merge_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to merge posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_merge_posts": False}
    )

    if not category_acl["can_merge_posts"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't merge posts in this category."
            )
        )
    if target.is_event:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "Events can't be merged.")
        )
    if (
        target.is_hidden
        and not category_acl["can_hide_posts"]
        and not target.is_first_post
    ):
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't merge posts the content you can't see."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't merge posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't merge posts in it.",
                )
            )


can_merge_post = return_boolean(allow_merge_post)


def allow_split_post(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to split posts.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_move_posts": False}
    )

    if not category_acl["can_move_posts"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't split posts in this category."
            )
        )
    if target.is_event:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "Events can't be split.")
        )
    if target.is_first_post:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "Thread's first post can't be split.")
        )
    if not category_acl["can_hide_posts"] and target.is_hidden:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't split posts the content you can't see."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't split posts in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't split posts in it.",
                )
            )


can_split_post = return_boolean(allow_split_post)


def allow_unhide_event(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to reveal events.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_events": 0}
    )

    if not category_acl["can_hide_events"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't reveal events in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't reveal events in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't reveal events in it.",
                )
            )


can_unhide_event = return_boolean(allow_unhide_event)


def allow_hide_event(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to hide events.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_events": 0}
    )

    if not category_acl["can_hide_events"]:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't hide events in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't hide events in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't hide events in it.",
                )
            )


can_hide_event = return_boolean(allow_hide_event)


def allow_delete_event(user_acl, target):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(
            pgettext_lazy("threads permission", "You have to sign in to delete events.")
        )

    category_acl = user_acl["categories"].get(
        target.category_id, {"can_hide_events": 0}
    )

    if category_acl["can_hide_events"] != 2:
        raise PermissionDenied(
            pgettext_lazy(
                "threads permission", "You can't delete events in this category."
            )
        )

    if not category_acl["can_close_threads"]:
        if target.category.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This category is closed. You can't delete events in it.",
                )
            )
        if target.thread.is_closed:
            raise PermissionDenied(
                pgettext_lazy(
                    "threads permission",
                    "This thread is closed. You can't delete events in it.",
                )
            )


can_delete_event = return_boolean(allow_delete_event)


def can_change_owned_thread(user_acl, target):
    if user_acl["is_anonymous"] or user_acl["user_id"] != target.starter_id:
        return False

    if target.category.is_closed or target.is_closed:
        return False

    return has_time_to_edit_thread(user_acl, target)


def has_time_to_edit_thread(user_acl, target):
    edit_time = (
        user_acl["categories"].get(target.category_id, {}).get("thread_edit_time", 0)
    )
    if edit_time:
        diff = timezone.now() - target.started_on
        diff_minutes = int(diff.total_seconds() / 60)
        return diff_minutes < edit_time

    return True


def has_time_to_edit_post(user_acl, target):
    edit_time = (
        user_acl["categories"].get(target.category_id, {}).get("post_edit_time", 0)
    )
    if edit_time:
        diff = timezone.now() - target.posted_on
        diff_minutes = int(diff.total_seconds() / 60)
        return diff_minutes < edit_time

    return True


def exclude_invisible_threads(
    user_acl, categories, queryset
):  # pylint: disable=too-many-branches
    show_all = []
    show_accepted_visible = []
    show_accepted = []
    show_visible = []
    show_owned = []
    show_owned_visible = []

    for category in categories:
        add_acl_to_obj(user_acl, category)

        if not (category.acl["can_see"] and category.acl["can_browse"]):
            continue

        can_hide = category.acl["can_hide_threads"]
        if category.acl["can_see_all_threads"]:
            can_mod = category.acl["can_approve_content"]

            if can_mod and can_hide:
                show_all.append(category)
            elif user_acl["is_authenticated"]:
                if not can_mod and not can_hide:
                    show_accepted_visible.append(category)
                elif not can_mod:
                    show_accepted.append(category)
                elif not can_hide:
                    show_visible.append(category)
            else:
                show_accepted_visible.append(category)
        elif user_acl["is_authenticated"]:
            if can_hide:
                show_owned.append(category)
            else:
                show_owned_visible.append(category)

    conditions = None
    if show_all:
        conditions = Q(category__in=show_all)

    if show_accepted_visible:
        if user_acl["is_authenticated"]:
            condition = Q(
                Q(starter_id=user_acl["user_id"]) | Q(is_unapproved=False),
                category__in=show_accepted_visible,
                is_hidden=False,
            )
        else:
            condition = Q(
                category__in=show_accepted_visible, is_hidden=False, is_unapproved=False
            )

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if show_accepted:
        condition = Q(
            Q(starter_id=user_acl["user_id"]) | Q(is_unapproved=False),
            category__in=show_accepted,
        )

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if show_visible:
        condition = Q(category__in=show_visible, is_hidden=False)

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if show_owned:
        condition = Q(category__in=show_owned, starter_id=user_acl["user_id"])

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if show_owned_visible:
        condition = Q(
            category__in=show_owned_visible,
            starter_id=user_acl["user_id"],
            is_hidden=False,
        )

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if not conditions:
        return Thread.objects.none()

    return queryset.filter(conditions)


def exclude_invisible_posts(user_acl, categories, queryset):
    if hasattr(categories, "__iter__"):
        return exclude_invisible_posts_in_categories(user_acl, categories, queryset)
    return exclude_invisible_posts_in_category(user_acl, categories, queryset)


def exclude_invisible_posts_in_categories(
    user_acl, categories, queryset
):  # pylint: disable=too-many-branches
    show_all = []
    show_approved = []
    show_approved_owned = []

    hide_invisible_events = []

    for category in categories:
        add_acl_to_obj(user_acl, category)

        if category.acl["can_approve_content"]:
            show_all.append(category.pk)
        else:
            if user_acl["is_authenticated"]:
                show_approved_owned.append(category.pk)
            else:
                show_approved.append(category.pk)

        if not category.acl["can_hide_events"]:
            hide_invisible_events.append(category.pk)

    conditions = None
    if show_all:
        conditions = Q(category__in=show_all)

    if show_approved:
        condition = Q(category__in=show_approved, is_unapproved=False)

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if show_approved_owned:
        condition = Q(
            Q(poster_id=user_acl["user_id"]) | Q(is_unapproved=False),
            category__in=show_approved_owned,
        )

        if conditions:
            conditions = conditions | condition
        else:
            conditions = condition

    if hide_invisible_events:
        queryset = queryset.exclude(
            category__in=hide_invisible_events, is_event=True, is_hidden=True
        )

    if not conditions:
        return Post.objects.none()

    return queryset.filter(conditions)


def exclude_invisible_posts_in_category(user_acl, category, queryset):
    add_acl_to_obj(user_acl, category)

    if not category.acl["can_approve_content"]:
        if user_acl["is_authenticated"]:
            queryset = queryset.filter(
                Q(is_unapproved=False) | Q(poster_id=user_acl["user_id"])
            )
        else:
            queryset = queryset.exclude(is_unapproved=True)

    if not category.acl["can_hide_events"]:
        queryset = queryset.exclude(is_event=True, is_hidden=True)

    return queryset



import pytest

from .models import Icon


@pytest.fixture
def favicon(db):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON, image="static/favicon.png", size=1, width=48, height=48
    )


@pytest.fixture
def favicon_32(db):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON_32,
        image="static/favicon-32.png",
        size=1,
        width=32,
        height=32,
    )


@pytest.fixture
def favicon_16(db):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON_16,
        image="static/favicon-16.png",
        size=1,
        width=16,
        height=16,
    )


@pytest.fixture
def apple_touch_icon(db):
    return Icon.objects.create(
        type=Icon.TYPE_APPLE_TOUCH_ICON,
        image="static/test.png",
        size=1,
        width=180,
        height=180,
    )



import pytest

from ..models import CacheVersion
from ..test import assert_invalidates_cache
from ..versions import invalidate_cache


def test_assertion_fails_if_specified_cache_is_not_invaldiated(cache_version):
    with pytest.raises(AssertionError):
        with assert_invalidates_cache(cache_version.cache):
            pass


def test_assertion_passess_if_specified_cache_is_invalidated(cache_version):
    with assert_invalidates_cache(cache_version.cache):
        invalidate_cache(cache_version.cache)


def test_assertion_fails_if_other_cache_is_invalidated(cache_version):
    CacheVersion.objects.create(cache="changed_test")
    with pytest.raises(AssertionError):
        with assert_invalidates_cache(cache_version.cache):
            invalidate_cache("changed_test")



import pytest
from django.conf import settings

from ..metadata import PluginsMetadataLoader


@pytest.fixture
def plugins_metadata():
    return PluginsMetadataLoader(settings.INSTALLED_PLUGINS)


def test_plugins_metadata_loader_returns_metadata_dict(plugins_metadata):
    metadata = plugins_metadata.get_metadata()
    assert metadata


def test_plugins_metadata_loader_caches_metadata(plugins_metadata):
    metadata = plugins_metadata.get_metadata()
    other_metadata = plugins_metadata.get_metadata()
    assert id(metadata) == id(other_metadata)


def test_plugins_metadata_loader_handles_plugins_without_manifests(plugins_metadata):
    metadata = plugins_metadata.get_metadata()
    assert "minimal_plugin" in metadata

    plugin_metadata = metadata["minimal_plugin"]
    assert plugin_metadata.package == "minimal_plugin"
    assert plugin_metadata.dirname == "minimal-plugin"
    assert plugin_metadata.has_manifest == False
    assert plugin_metadata.manifest_error is None
    assert plugin_metadata.name is None


def test_plugins_metadata_loader_handles_plugins_with_empty_manifests(plugins_metadata):
    metadata = plugins_metadata.get_metadata()
    assert "empty_manifest_plugin" in metadata

    plugin_metadata = metadata["empty_manifest_plugin"]
    assert plugin_metadata.package == "empty_manifest_plugin"
    assert plugin_metadata.dirname == "empty-manifest-plugin"
    assert plugin_metadata.has_manifest == True
    assert plugin_metadata.manifest_error is None
    assert plugin_metadata.name is None


def test_plugins_metadata_loader_handles_plugins_with_full_manifests(plugins_metadata):
    metadata = plugins_metadata.get_metadata()
    assert "full_manifest_plugin" in metadata

    plugin_metadata = metadata["full_manifest_plugin"]
    assert plugin_metadata.package == "full_manifest_plugin"
    assert plugin_metadata.dirname == "full-manifest-plugin"
    assert plugin_metadata.has_manifest == True
    assert plugin_metadata.manifest_error is None
    assert plugin_metadata.name == "Example plugin with complete manifest"


def test_plugins_metadata_loader_handles_plugins_with_invalid_manifests(
    plugins_metadata,
):
    metadata = plugins_metadata.get_metadata()
    assert "invalid_manifest_plugin" in metadata

    plugin_metadata = metadata["invalid_manifest_plugin"]
    assert plugin_metadata.package == "invalid_manifest_plugin"
    assert plugin_metadata.dirname == "invalid-manifest-plugin"
    assert plugin_metadata.has_manifest == False
    assert plugin_metadata.manifest_error is not None
    assert plugin_metadata.name is None


def test_plugins_metadata_loader_handles_empty_plugins_list():
    plugins_metadata = PluginsMetadataLoader([])
    metadata = plugins_metadata.get_metadata()
    assert metadata == {}
    assert id(metadata) == id(metadata)



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ....conf.shortcuts import get_dynamic_settings
from ....core.pgutils import chunk_queryset
from ...deletesrecord import record_user_deleted_by_self
from ...permissions import can_delete_own_account

User = get_user_model()


class Command(BaseCommand):
    help = (
        "Deletes accounts of users that have requested it. "
        "Leaves their content behind, but anonymises it."
    )

    def handle(self, *args, **options):
        users_deleted = 0
        settings = get_dynamic_settings()

        queryset = User.objects.filter(is_deleting_account=True)

        for user in chunk_queryset(queryset):
            if can_delete_own_account(settings, user, user):
                user.delete(anonymous_username=settings.anonymous_username)
                record_user_deleted_by_self()
                users_deleted += 1

        self.stdout.write("Deleted users: %s" % users_deleted)



from django.urls import reverse

from ...test import assert_contains
from ..auth import is_admin_authorized

admin_link = reverse("misago:admin:index")


def test_login_form_is_displayed(db, client):
    response = client.get(admin_link)
    assert response.status_code == 200
    assert_contains(response, "Administration")
    assert_contains(response, "Sign in")
    assert_contains(response, "Username or e-mail")
    assert_contains(response, "Password")


def test_attempt_to_login_using_invalid_credentials_fails(db, client):
    response = client.post(admin_link, {"username": "no", "password": "no"})
    assert_contains(response, "Login or password is incorrect.")


def test_attempt_to_login_using_invalid_password_fails(client, superuser):
    response = client.post(
        admin_link, {"username": superuser.username, "password": "no"}
    )
    assert_contains(response, "Login or password is incorrect.")


def test_attempt_to_login_without_staff_status_fails(client, user, user_password):
    response = client.post(
        admin_link, {"username": user.username, "password": user_password}
    )
    assert_contains(response, "Your account does not have admin privileges.")


def test_attempt_to_login_as_superuser_without_staff_status_fails(
    client, user, user_password
):
    user.is_superuser = True
    user.save()

    response = client.post(
        admin_link, {"username": user.username, "password": user_password}
    )
    assert_contains(response, "Your account does not have admin privileges.")


def test_user_with_staff_status_is_logged_to_admin(client, staffuser, user_password):
    response = client.post(
        admin_link, {"username": staffuser.username, "password": user_password}
    )
    assert is_admin_authorized(response.wsgi_request)
    assert response.wsgi_request.user == staffuser


def test_login_form_redirects_user_to_admin_index_after_successful_login(
    client, staffuser, user_password
):
    response = client.post(
        admin_link, {"username": staffuser.username, "password": user_password}
    )
    assert response["location"] == admin_link



# Generated by Django 1.11.13 on 2018-06-24 00:13
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import misago.users.models.datadownload


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0013_auto_20180609_1523")]

    operations = [
        migrations.CreateModel(
            name="DataDownload",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.PositiveIntegerField(
                        choices=[
                            (0, "Pending"),
                            (1, "Processing"),
                            (2, "Ready"),
                            (3, "Expired"),
                        ],
                        db_index=True,
                        default=0,
                    ),
                ),
                ("requester_name", models.CharField(max_length=255)),
                (
                    "requested_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("expires_on", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "file",
                    models.FileField(
                        blank=True,
                        max_length=255,
                        null=True,
                        upload_to=misago.users.models.datadownload.get_data_upload_to,
                    ),
                ),
                (
                    "requester",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"ordering": ["-pk"]},
        )
    ]



from django.http import HttpRequest
from django.shortcuts import redirect


def redirect_subscribed_to_watched(request: HttpRequest, **kwargs):
    return redirect(
        request.path.replace("/subscribed/", "/watched/"),
        permanent=True,
    )



from django.urls import reverse


def test_disable_email_notifications_view_disables_emails_for_watched_thread(
    watched_thread_factory, user, user_client, thread
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    response = user_client.get(
        reverse(
            "misago:notifications-disable-email",
            kwargs={
                "watched_thread_id": watched_thread.id,
                "secret": watched_thread.secret,
            },
        )
    )
    assert response.status_code == 200

    watched_thread.refresh_from_db()
    assert not watched_thread.send_emails


def test_disable_email_notifications_view_works_without_authentication(
    watched_thread_factory, user, client, thread
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    response = client.get(
        reverse(
            "misago:notifications-disable-email",
            kwargs={
                "watched_thread_id": watched_thread.id,
                "secret": watched_thread.secret,
            },
        )
    )
    assert response.status_code == 200

    watched_thread.refresh_from_db()
    assert not watched_thread.send_emails


def test_disable_email_notifications_view_returns_404_if_secret_is_invalid(
    watched_thread_factory, user, client, thread
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    response = client.get(
        reverse(
            "misago:notifications-disable-email",
            kwargs={
                "watched_thread_id": watched_thread.id,
                "secret": "invalid",
            },
        )
    )
    assert response.status_code == 404


def test_disable_email_notifications_view_returns_404_if_watched_thread_is_not_found(
    watched_thread_factory, user, client, thread
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    response = client.get(
        reverse(
            "misago:notifications-disable-email",
            kwargs={
                "watched_thread_id": watched_thread.id + 1,
                "secret": watched_thread.secret,
            },
        )
    )
    assert response.status_code == 404



from ..parser import parse


def test_parser_converts_unmarked_links_to_hrefs(request_mock, user, snapshot):
    text = "Lorem ipsum http://test.com"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_parser_skips_links_in_inline_code_markdown(request_mock, user, snapshot):
    text = "Lorem ipsum `http://test.com`"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_parser_skips_links_in_inline_code_bbcode(request_mock, user, snapshot):
    text = "Lorem ipsum [code]http://test.com[/code]"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_parser_skips_links_in_code_bbcode(request_mock, user, snapshot):
    text = """
[code]
http://test.com
[/code]
    """
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_absolute_link_to_site_is_changed_to_relative_link(
    request_mock, user, snapshot
):
    text = "clean_links step cleans http://example.com"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_absolute_link_to_site_is_added_to_internal_links_list(request_mock, user):
    text = "clean_links step cleans http://example.com"
    result = parse(text, request_mock, user)
    assert result["internal_links"] == ["/"]


def test_absolute_link_to_site_without_schema_is_changed_to_relative_link(
    request_mock, user, snapshot
):
    text = "clean_links step cleans example.com"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_absolute_link_to_site_without_schema_is_added_to_internal_links_list(
    request_mock, user
):
    text = "clean_links step cleans example.com"
    result = parse(text, request_mock, user)
    assert result["internal_links"] == ["/"]


def test_absolute_link_with_path_to_site_is_changed_to_relative_link(
    request_mock, user, snapshot
):
    text = "clean_links step cleans http://example.com/somewhere-something/"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_absolute_link_with_path_to_site_is_added_to_internal_links_list(
    request_mock, user
):
    text = "clean_links step cleans http://example.com/somewhere-something/"
    result = parse(text, request_mock, user)
    assert result["internal_links"] == ["/somewhere-something/"]


def test_full_link_with_path_text_is_set_to_domain_and_path(request_mock, user):
    text = "clean_links step cleans http://example.com/somewhere-something/"
    result = parse(text, request_mock, user)
    assert ">example.com/somewhere-something/<" in result["parsed_text"]


def test_outgoing_link_is_added_to_outgoing_links_list(request_mock, user):
    text = "clean_links step cleans https://other.com"
    result = parse(text, request_mock, user)
    assert result["outgoing_links"] == ["other.com"]


def test_outgoing_llink_includes_external_nofollow_and_noopener(request_mock, user):
    text = "Lorem [url]https://dummyimage.com/g/1200/500[/url] ipsum"
    result = parse(text, request_mock, user)
    assert 'rel="external nofollow noopener"' in result["parsed_text"]


def test_outgoing_link_without_scheme_is_added_to_outgoing_links_list(
    request_mock, user
):
    text = "clean_links step cleans other.com"
    result = parse(text, request_mock, user)
    assert result["outgoing_links"] == ["other.com"]


def test_outgoing_link_with_path_is_added_to_outgoing_links_list(request_mock, user):
    text = "clean_links step cleans other.com/some/path/"
    result = parse(text, request_mock, user)
    assert result["outgoing_links"] == ["other.com/some/path/"]


def test_local_image_is_changed_to_relative_link(request_mock, user, snapshot):
    text = "clean_links step cleans !(example.com/media/img.png)"
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_local_image_is_added_to_images_list(request_mock, user):
    text = "clean_links step cleans !(example.com/media/img.png)"
    result = parse(text, request_mock, user)
    assert result["images"] == ["/media/img.png"]


def test_remote_image_is_added_to_images_list(request_mock, user):
    text = "clean_links step cleans !(other.com/media/img.png)"
    result = parse(text, request_mock, user)
    assert result["images"] == ["other.com/media/img.png"]


def test_local_image_link_is_added_to_images_and_links_lists(request_mock, user):
    text = "clean_links step cleans [!(example.com/media/img.png)](example.com/test/)"
    result = parse(text, request_mock, user)
    assert result["internal_links"] == ["/test/"]
    assert result["images"] == ["/media/img.png"]


def test_remote_image_link_is_added_to_images_and_links_lists(request_mock, user):
    text = "clean_links step cleans [!(other.com/media/img.png)](other.com/test/)"
    result = parse(text, request_mock, user)
    assert result["outgoing_links"] == ["other.com/test/"]
    assert result["images"] == ["other.com/media/img.png"]


def test_parser_adds_shva_to_attachment_link_querystring_if_force_option_is_enabled(
    request_mock, user
):
    text = "clean_links step cleans ![3.png](http://example.com/a/thumb/test/43/)"
    result = parse(text, request_mock, user, force_shva=True)
    assert "/a/thumb/test/43/?shva=1" in result["parsed_text"]


def test_parser_skips_shva_in_attachment_link_querystring_if_force_option_is_omitted(
    request_mock, user
):
    text = "clean_links step cleans ![3.png](http://example.com/a/thumb/test/43/)"
    result = parse(text, request_mock, user)
    assert "?shva=1" not in result["parsed_text"]



from django.contrib.auth import get_user_model
from django.template.context_processors import static
from rest_framework.decorators import api_view
from rest_framework.response import Response

from ...conf import settings

User = get_user_model()


@api_view()
def mention_suggestions(request):
    suggestions = []

    query = request.query_params.get("q", "").lower().strip()[:100]
    if query:
        queryset = User.objects.filter(slug__startswith=query, is_active=True).order_by(
            "slug"
        )[:10]

        for user in queryset:
            try:
                avatar = user.avatars[-1]["url"]
            except IndexError:
                avatar = request.settings.blank_avatar
                if not avatar:
                    avatar = static(settings.MISAGO_BLANK_AVATAR)

            suggestions.append({"username": user.username, "avatar": avatar})

    return Response(suggestions)



from typing import List

from django.core.cache import cache
from django.urls import reverse

from ..acl import ACL_CACHE
from ..users.models import AnonymousUser
from .models import Category

CACHE_NAME = "categories_map"


def get_categories_map(request) -> List[dict]:
    cache_key = get_cache_key(request)
    categories_map = cache.get(cache_key, None)
    if categories_map is None:
        categories_map = get_categories_map_from_db(request)
        cache.set(cache_key, categories_map)
    return categories_map


def get_cache_key(request) -> str:
    acl_version = request.cache_versions[ACL_CACHE]
    if request.user.is_authenticated:
        user_acl_key = request.user.acl_key
    else:
        user_acl_key = AnonymousUser.acl_key

    return f"{CACHE_NAME}:{acl_version}:{user_acl_key}"


MAP_CATEGORY_FIELDS = (
    "id",
    "name",
    "slug",
    "short_name",
    "color",
    "level",
)


def get_categories_map_from_db(request) -> List[dict]:
    categories_list = []
    queryset = Category.objects.filter(
        id__in=request.user_acl["visible_categories"],
        level=1,
        special_role__isnull=True,
    ).order_by("lft")

    for category in queryset.values(*MAP_CATEGORY_FIELDS):
        category_url = reverse(
            "misago:category",
            kwargs={"pk": category["id"], "slug": category["slug"]},
        )

        category_json = {
            "id": category["id"],
            "name": category["name"],
            "shortName": category["short_name"],
            "color": category["color"],
            "url": category_url,
        }

        categories_list.append(category_json)

    return categories_list



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...core.utils import encode_json_html
from ..models import Ban
from ..test import create_test_user
from ..tokens import make_activation_token

User = get_user_model()


class ActivationViewsTests(TestCase):
    def test_request_view_returns_200(self):
        """request new activation link view returns 200"""
        response = self.client.get(reverse("misago:request-activation"))
        self.assertEqual(response.status_code, 200)

    def test_view_activate_banned(self):
        """activate banned user shows error"""
        user = create_test_user("User", "user@example.com", requires_activation=1)
        activation_token = make_activation_token(user)

        Ban.objects.create(
            check_type=Ban.USERNAME, banned_value="user", user_message="Nope!"
        )

        response = self.client.get(
            reverse(
                "misago:activate-by-token",
                kwargs={"pk": user.pk, "token": activation_token},
            )
        )
        self.assertContains(response, encode_json_html("<p>Nope!</p>"), status_code=403)

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.requires_activation, 1)

    def test_view_activate_invalid_token(self):
        """activate with invalid token shows error"""
        user = create_test_user("User", "user@example.com", requires_activation=1)
        activation_token = make_activation_token(user)

        response = self.client.get(
            reverse(
                "misago:activate-by-token",
                kwargs={"pk": user.pk, "token": activation_token + "acd"},
            )
        )
        self.assertEqual(response.status_code, 400)

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.requires_activation, 1)

    def test_view_activate_disabled(self):
        """activate disabled user shows error"""
        user = create_test_user("User", "user@example.com", is_active=False)
        activation_token = make_activation_token(user)

        response = self.client.get(
            reverse(
                "misago:activate-by-token",
                kwargs={"pk": user.pk, "token": activation_token},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_view_activate_active(self):
        """activate active user shows error"""
        user = create_test_user("User", "user@example.com")
        activation_token = make_activation_token(user)

        response = self.client.get(
            reverse(
                "misago:activate-by-token",
                kwargs={"pk": user.pk, "token": activation_token},
            )
        )
        self.assertEqual(response.status_code, 200)

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.requires_activation, 0)

    def test_view_activate_inactive(self):
        """activate inactive user passess"""
        user = create_test_user("User", "user@example.com", requires_activation=1)
        activation_token = make_activation_token(user)

        response = self.client.get(
            reverse(
                "misago:activate-by-token",
                kwargs={"pk": user.pk, "token": activation_token},
            )
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "your account has been activated!")

        user = User.objects.get(pk=user.pk)
        self.assertEqual(user.requires_activation, 0)


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_request_activatiom_view_returns_403_if_oauth_is_enabled(db, client):
    response = client.get(reverse("misago:request-activation"))
    assert response.status_code == 403


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_activate_with_token_view_returns_403_if_oauth_is_enabled(db, client):
    user = create_test_user("User", "user@example.com", requires_activation=1)
    activation_token = make_activation_token(user)

    response = client.post(
        reverse(
            "misago:activate-by-token",
            kwargs={"pk": user.pk, "token": activation_token},
        )
    )
    assert response.status_code == 403



from datetime import timedelta
from io import StringIO

import pytest
from django.core import management
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..management.commands import clearattachments
from ..models import Attachment, AttachmentType


@pytest.fixture
def attachment_type(db):
    return AttachmentType.objects.order_by("id").last()


def create_attachment(attachment_type, uploaded_on, post=None):
    return Attachment.objects.create(
        secret=Attachment.generate_new_secret(),
        post=post,
        filetype=attachment_type,
        size=1000,
        uploaded_on=uploaded_on,
        uploader_name="User",
        uploader_slug="user",
        filename="testfile_%s.zip" % (Attachment.objects.count() + 1),
    )


def call_command():
    command = clearattachments.Command()

    out = StringIO()
    management.call_command(command, stdout=out)
    return out.getvalue().strip().splitlines()[-1].strip()


def test_command_works_if_there_are_no_attachments(db):
    command_output = call_command()
    assert command_output == "No unused attachments were cleared"


@override_dynamic_settings(unused_attachments_lifetime=2)
def test_recent_attachment_is_not_cleared(attachment_type):
    attachment = create_attachment(attachment_type, timezone.now())
    command_output = call_command()
    assert command_output == "No unused attachments were cleared"


@override_dynamic_settings(unused_attachments_lifetime=2)
def test_old_used_attachment_is_not_cleared(attachment_type, post):
    uploaded_on = timezone.now() - timedelta(hours=3)
    attachment = create_attachment(attachment_type, uploaded_on, post)
    command_output = call_command()
    assert command_output == "No unused attachments were cleared"


@override_dynamic_settings(unused_attachments_lifetime=2)
def test_old_unused_attachment_is_cleared(attachment_type):
    uploaded_on = timezone.now() - timedelta(hours=3)
    attachment = create_attachment(attachment_type, uploaded_on)
    command_output = call_command()
    assert command_output == "Cleared 1 attachments"

    with pytest.raises(Attachment.DoesNotExist):
        attachment.refresh_from_db()



import pytest


@pytest.fixture
def notify_on_new_private_thread_mock(mocker):
    return mocker.patch("misago.threads.participants.notify_on_new_private_thread")


@pytest.fixture
def notify_on_new_thread_reply_mock(mocker):
    return mocker.patch(
        "misago.threads.api.postingendpoint.notifications.notify_on_new_thread_reply"
    )



from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import Http404
from django.utils.translation import pgettext
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from ...acl.objectacl import add_acl_to_obj
from ...core.shortcuts import get_int_or_404
from ...users.audittrail import create_audit_trail
from ..models import Poll
from ..permissions import (
    allow_delete_poll,
    allow_edit_poll,
    allow_see_poll_votes,
    allow_start_poll,
    can_start_poll,
)
from ..serializers import (
    EditPollSerializer,
    NewPollSerializer,
    PollSerializer,
    PollVoteSerializer,
)
from ..viewmodels import ForumThread
from .pollvotecreateendpoint import poll_vote_create


class ViewSet(viewsets.ViewSet):
    thread = None

    def get_thread(self, request, thread_pk):
        return self.thread(  # pylint: disable=not-callable
            request, get_int_or_404(thread_pk)
        ).unwrap()

    def get_poll(self, thread, pk):
        try:
            poll_id = get_int_or_404(pk)
            if thread.poll.pk != poll_id:
                raise Http404()

            poll = Poll.objects.get(pk=thread.poll.pk)

            poll.thread = thread
            poll.category = thread.category

            return poll
        except Poll.DoesNotExist:
            raise Http404()

    @transaction.atomic
    def create(self, request, thread_pk):
        thread = self.get_thread(request, thread_pk)
        allow_start_poll(request.user_acl, thread)

        try:
            if thread.poll and thread.poll.pk:
                raise PermissionDenied(
                    pgettext("polls api", "There's already a poll in this thread.")
                )
        except Poll.DoesNotExist:
            pass

        instance = Poll(
            thread=thread,
            category=thread.category,
            poster=request.user,
            poster_name=request.user.username,
            poster_slug=request.user.slug,
        )

        serializer = NewPollSerializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.save()

        add_acl_to_obj(request.user_acl, instance)
        for choice in instance.choices:
            choice["selected"] = False

        thread.has_poll = True
        thread.save()

        create_audit_trail(request, instance)

        return Response(PollSerializer(instance).data)

    @transaction.atomic
    def update(self, request, thread_pk, pk=None):
        thread = self.get_thread(request, thread_pk)
        instance = self.get_poll(thread, pk)

        allow_edit_poll(request.user_acl, instance)

        serializer = EditPollSerializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.save()

        add_acl_to_obj(request.user_acl, instance)
        instance.make_choices_votes_aware(request.user)

        create_audit_trail(request, instance)

        return Response(PollSerializer(instance).data)

    @transaction.atomic
    def delete(self, request, thread_pk, pk=None):
        thread = self.get_thread(request, thread_pk)
        instance = self.get_poll(thread, pk)

        allow_delete_poll(request.user_acl, instance)

        thread.poll.delete()

        thread.has_poll = False
        thread.save()

        return Response({"can_start_poll": can_start_poll(request.user_acl, thread)})

    @action(detail=True, methods=["get", "post"])
    def votes(self, request, thread_pk, pk=None):
        if request.method == "POST":
            return self.post_votes(request, thread_pk, pk)
        return self.get_votes(request, thread_pk, pk)

    @transaction.atomic
    def post_votes(self, request, thread_pk, pk=None):
        thread = self.get_thread(request, thread_pk)
        instance = self.get_poll(thread, pk)

        return poll_vote_create(request, thread, instance)

    def get_votes(self, request, thread_pk, pk=None):
        poll_pk = get_int_or_404(pk)

        try:
            thread = self.get_thread(request, thread_pk)
            if thread.poll.pk != poll_pk:
                raise Http404()
        except Poll.DoesNotExist:
            raise Http404()

        allow_see_poll_votes(request.user_acl, thread.poll)

        choices = []
        voters = {}

        for choice in thread.poll.choices:
            choice["voters"] = []
            voters[choice["hash"]] = choice["voters"]

            choices.append(choice)

        queryset = thread.poll.pollvote_set.values(
            "voter_id", "voter_name", "voter_slug", "voted_on", "choice_hash"
        )

        for voter in queryset.order_by("voter_name").iterator():
            voters[voter["choice_hash"]].append(PollVoteSerializer(voter).data)

        return Response(choices)


class ThreadPollViewSet(ViewSet):
    thread = ForumThread



from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext_lazy

__all__ = ["authenticated_only", "anonymous_only"]


def authenticated_only(f):
    def perm_decorator(user_acl, target):
        if user_acl["is_anonymous"]:
            raise PermissionDenied(
                pgettext_lazy(
                    "view decorator", "You have to sign in to perform this action."
                )
            )
        return f(user_acl, target)

    return perm_decorator


def anonymous_only(f):
    def perm_decorator(user_acl, target):
        if user_acl["is_authenticated"]:
            raise PermissionDenied(
                pgettext_lazy("view decorator", "Only guests can perform this action.")
            )
        return f(user_acl, target)

    return perm_decorator



from django.contrib.auth import get_user_model, logout
from django.contrib.auth.password_validation import validate_password
from django.utils.translation import pgettext
from rest_framework import serializers

from ..online.tracker import clear_tracking
from ..permissions import allow_delete_own_account
from ..validators import validate_email, validate_username

User = get_user_model()

__all__ = [
    "ForumOptionsSerializer",
    "EditSignatureSerializer",
    "ChangeUsernameSerializer",
    "ChangePasswordSerializer",
    "ChangeEmailSerializer",
    "DeleteOwnAccountSerializer",
]


class ForumOptionsSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = [
            "is_hiding_presence",
            "limits_private_thread_invites_to",
            "watch_started_threads",
            "watch_replied_threads",
            "watch_new_private_threads_by_followed",
            "watch_new_private_threads_by_other_users",
            "notify_new_private_threads_by_followed",
            "notify_new_private_threads_by_other_users",
        ]
        extra_kwargs = {
            "limits_private_thread_invites_to": {"required": True},
            "watch_started_threads": {"required": True},
            "watch_replied_threads": {"required": True},
            "watch_new_private_threads_by_followed": {"required": True},
            "watch_new_private_threads_by_other_users": {"required": True},
            "notify_new_private_threads_by_followed": {"required": True},
            "notify_new_private_threads_by_other_users": {"required": True},
        }


class EditSignatureSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["signature"]

    def validate(self, data):
        settings = self.context["settings"]
        if len(data.get("signature", "")) > settings.signature_length_max:
            raise serializers.ValidationError(
                pgettext("edit signature serializer", "Signature is too long.")
            )

        return data


class ChangeUsernameSerializer(serializers.Serializer):
    username = serializers.CharField(max_length=200, required=False, allow_blank=True)

    def validate(self, data):
        username = data.get("username")
        if not username:
            raise serializers.ValidationError(
                pgettext("change username serializer", "Enter new username.")
            )

        user = self.context["user"]
        if username == user.username:
            raise serializers.ValidationError(
                pgettext(
                    "change username serializer", "New username is same as current one."
                )
            )

        settings = self.context["settings"]
        validate_username(settings, username)

        return data

    def change_username(self, changed_by):
        user = self.context["user"]
        user.set_username(self.validated_data["username"], changed_by=changed_by)
        user.save(update_fields=["username", "slug"])


class ChangePasswordSerializer(serializers.Serializer):
    password = serializers.CharField(max_length=200, trim_whitespace=False)
    new_password = serializers.CharField(max_length=200, trim_whitespace=False)

    def validate_password(self, value):
        if not self.context["user"].check_password(value):
            raise serializers.ValidationError(
                pgettext("change password serializer", "Entered password is invalid.")
            )
        return value

    def validate_new_password(self, value):
        validate_password(value, user=self.context["user"])
        return value


class ChangeEmailSerializer(serializers.Serializer):
    password = serializers.CharField(max_length=200, trim_whitespace=False)
    new_email = serializers.CharField(max_length=200)

    def validate_password(self, value):
        if not self.context["user"].check_password(value):
            raise serializers.ValidationError(
                pgettext("change email serializer", "Entered password is invalid.")
            )
        return value

    def validate_new_email(self, value):
        if not value:
            raise serializers.ValidationError(
                pgettext(
                    "change email serializer", "You have to enter new e-mail address."
                )
            )

        if value.lower() == self.context["user"].email.lower():
            raise serializers.ValidationError(
                pgettext(
                    "change email serializer", "New e-mail is same as current one."
                )
            )

        validate_email(value)

        return value


class DeleteOwnAccountSerializer(serializers.Serializer):
    password = serializers.CharField(max_length=200, trim_whitespace=False)

    def validate_password(self, value):
        if not self.context["user"].check_password(value):
            raise serializers.ValidationError(
                pgettext(
                    "delete own account serializer", "Entered password is invalid."
                )
            )
        return value

    def mark_account_for_deletion(self, request):
        """
        Deleting user account can be costful, so just mark account for deletion,
        deactivate it and sign user out.
        """
        profile = self.context["user"]
        allow_delete_own_account(request.settings, request.user, profile)

        logout(request)
        clear_tracking(request)

        profile.mark_for_delete()



from ..models import Setting


def test_setting_value_is_hydrated_by_getter(db):
    setting = Setting(python_type="list", dry_value="lorem,ipsum")
    assert setting.value == ["lorem", "ipsum"]


def test_setting_value_is_dehydrated_by_setter(db):
    setting = Setting(python_type="list")
    setting.value = ["lorem", "ipsum"]
    assert setting.dry_value == "lorem,ipsum"


def test_setting_value_is_set_to_none(db):
    setting = Setting(python_type="list", dry_value="lorem,ipsum")
    setting.value = None
    assert setting.dry_value is None



from django.template.defaultfilters import slugify as django_slugify
from unidecode import unidecode


def default(string):
    string = str(string)
    string = unidecode(string)
    return django_slugify(string.replace("_", " ").strip())



from django.urls import reverse

from ...users.test import AuthenticatedUserTestCase
from ..models import Agreement


class RequiredAgreementTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:index")

        Agreement.objects.invalidate_cache()

    def tearDown(self):
        Agreement.objects.invalidate_cache()

    def test_tos_link(self):
        Agreement.objects.create(
            type=Agreement.TYPE_TOS, link="https://test-agreement.com", is_active=True
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_tos_text(self):
        Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=True
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_tos_text_and_link(self):
        Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            link="https://test-agreement.com",
            text="Lorem ipsum",
            is_active=True,
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_privacy_link(self):
        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="https://test-agreement.com",
            is_active=True,
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_privacy_text(self):
        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY, text="Lorem ipsum", is_active=True
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_privacy_text_and_link(self):
        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="https://test-agreement.com",
            text="Lorem ipsum",
            is_active=True,
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

    def test_both(self):
        Agreement.objects.create(
            type=Agreement.TYPE_TOS,
            link="https://test-agreement.com",
            text="Lorem ipsum",
            is_active=True,
        )

        Agreement.objects.create(
            type=Agreement.TYPE_PRIVACY,
            link="https://test-agreement.com",
            text="Lorem ipsum",
            is_active=True,
        )

        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)



def test_threads_subscribed_list_redirects_to_watched(db, client):
    response = client.get("/subscribed/")
    assert response.status_code == 301
    assert response.headers["location"] == "/watched/"


def test_category_subscribed_list_redirects_to_watched(db, client, default_category):
    response = client.get(default_category.get_absolute_url() + "subscribed/")
    assert response.status_code == 301
    assert (
        response.headers["location"] == default_category.get_absolute_url() + "watched/"
    )


def test_private_threads_subscribed_list_redirects_to_watched(db, client):
    response = client.get("/private-threads/subscribed/")
    assert response.status_code == 301
    assert response.headers["location"] == "/private-threads/watched/"



from django.urls import reverse
from rest_framework import serializers

from ..models import Attachment

__all__ = ["AttachmentSerializer"]


class AttachmentSerializer(serializers.ModelSerializer):
    post = serializers.PrimaryKeyRelatedField(read_only=True)

    acl = serializers.SerializerMethodField()
    is_image = serializers.SerializerMethodField()
    filetype = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        model = Attachment
        fields = [
            "id",
            "filetype",
            "post",
            "uploaded_on",
            "uploader_name",
            "filename",
            "size",
            "acl",
            "is_image",
            "url",
        ]

    def get_acl(self, obj):
        try:
            return obj.acl
        except AttributeError:
            return None

    def get_is_image(self, obj):
        return obj.is_image

    def get_filetype(self, obj):
        return obj.filetype.name

    def get_url(self, obj):
        return {
            "index": obj.get_absolute_url(),
            "thumb": obj.get_thumbnail_url(),
            "uploader": self.get_uploader_url(obj),
        }

    def get_uploader_url(self, obj):
        if obj.uploader_id:
            return reverse(
                "misago:user", kwargs={"slug": obj.uploader_slug, "pk": obj.uploader_id}
            )



from unittest.mock import Mock

import pytest

from ..poststracker import save_read


@pytest.fixture
def read_thread(user, thread):
    save_read(user, thread.first_post)
    return thread


@pytest.fixture
def anonymous_request_mock(dynamic_settings, anonymous_user, anonymous_user_acl):
    return Mock(
        settings=dynamic_settings, user=anonymous_user, user_acl=anonymous_user_acl
    )


@pytest.fixture
def request_mock(dynamic_settings, user, user_acl):
    return Mock(settings=dynamic_settings, user=user, user_acl=user_acl)



EQUAL = 0
CONTAINS = 1
STARTS_WITH = 2
ENDS_WITH = 3


def filter_queryset(queryset, attr, search, *, case_sensitive=False):
    mode = get_mode(search)
    search = search.strip("*")

    if not search:
        return queryset

    queryset_filter = get_queryset_filter(
        attr, mode, search, case_sensitive=case_sensitive
    )
    return queryset.filter(**queryset_filter)


def get_mode(search):
    if search.startswith("*") and search.endswith("*"):
        return CONTAINS
    if search.endswith("*"):
        return STARTS_WITH
    if search.startswith("*"):
        return ENDS_WITH
    return EQUAL


def get_queryset_filter(attr, mode, search, *, case_sensitive=False):
    if mode is STARTS_WITH:
        if case_sensitive:
            return {"%s__startswith" % attr: search}
        return {"%s__istartswith" % attr: search}

    if mode is ENDS_WITH:
        if case_sensitive:
            return {"%s__endswith" % attr: search}
        return {"%s__iendswith" % attr: search}

    if mode is CONTAINS:
        if case_sensitive:
            return {"%s__contains" % attr: search}
        return {"%s__icontains" % attr: search}

    if case_sensitive:
        return {attr: search}

    return {"%s__iexact" % attr: search}



# Generated by Django 1.11.17 on 2019-01-03 21:15
from django.db import migrations, models
import django.db.models.deletion
import misago.themes.uploadto
import mptt.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Css",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("url", models.URLField(blank=True, max_length=255, null=True)),
                (
                    "source_file",
                    models.FileField(
                        blank=True,
                        max_length=255,
                        null=True,
                        upload_to=misago.themes.uploadto.upload_source_css_to,
                    ),
                ),
                ("source_hash", models.CharField(blank=True, max_length=8, null=True)),
                ("source_needs_building", models.BooleanField(default=False)),
                (
                    "build_file",
                    models.FileField(
                        blank=True,
                        max_length=255,
                        null=True,
                        upload_to=misago.themes.uploadto.upload_build_css_to,
                    ),
                ),
                ("build_hash", models.CharField(blank=True, max_length=8, null=True)),
                ("size", models.PositiveIntegerField(default=0)),
                ("order", models.IntegerField(default=0)),
                ("modified_on", models.DateTimeField(auto_now=True)),
            ],
            options={"ordering": ["order"]},
        ),
        migrations.CreateModel(
            name="Media",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "file",
                    models.ImageField(
                        max_length=255, upload_to=misago.themes.uploadto.upload_media_to
                    ),
                ),
                ("hash", models.CharField(max_length=8)),
                ("type", models.CharField(max_length=255)),
                ("width", models.PositiveIntegerField(default=0)),
                ("height", models.PositiveIntegerField(default=0)),
                ("size", models.PositiveIntegerField()),
                (
                    "thumbnail",
                    models.ImageField(
                        blank=True,
                        max_length=255,
                        null=True,
                        upload_to=misago.themes.uploadto.upload_media_thumbnail_to,
                    ),
                ),
                ("modified_on", models.DateTimeField(auto_now=True)),
            ],
            options={"ordering": ["name"]},
        ),
        migrations.CreateModel(
            name="Theme",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "dirname",
                    models.CharField(
                        default=misago.themes.uploadto.generate_theme_dirname,
                        max_length=8,
                    ),
                ),
                ("is_default", models.BooleanField(default=False)),
                ("is_active", models.BooleanField(default=False)),
                ("version", models.CharField(blank=True, max_length=255, null=True)),
                ("author", models.CharField(blank=True, max_length=255, null=True)),
                ("url", models.URLField(blank=True, max_length=255, null=True)),
                ("lft", models.PositiveIntegerField(db_index=True, editable=False)),
                ("rght", models.PositiveIntegerField(db_index=True, editable=False)),
                ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
                ("level", models.PositiveIntegerField(db_index=True, editable=False)),
                (
                    "parent",
                    mptt.fields.TreeForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="children",
                        to="misago_themes.Theme",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AddField(
            model_name="media",
            name="theme",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="media",
                to="misago_themes.Theme",
            ),
        ),
        migrations.AddField(
            model_name="css",
            name="theme",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="css",
                to="misago_themes.Theme",
            ),
        ),
    ]



from django.utils.functional import SimpleLazyObject

from .dynamicsettings import DynamicSettings


def dynamic_settings_middleware(get_response):
    """Sets request.settings attribute with DynamicSettings."""

    def middleware(request):
        def get_dynamic_settings():
            return DynamicSettings(request.cache_versions)

        request.settings = SimpleLazyObject(get_dynamic_settings)
        return get_response(request)

    return middleware



import json
import os
import shutil
from tempfile import TemporaryDirectory

from django.http import FileResponse

from ...core.utils import slugify


def export_theme(theme):
    with TemporaryDirectory() as tmp_dir:
        export_dir = create_export_directory(tmp_dir, theme)

        manifest = create_theme_manifest(theme)

        manifest["css"] = write_theme_css(export_dir, theme)
        manifest["media"] = write_theme_media(export_dir, theme)
        write_theme_manifest(export_dir, manifest)

        export_file = zip_theme_export(tmp_dir, export_dir)
        export_filename = os.path.split(export_file)[-1]

        response = FileResponse(open(export_file, "rb"), content_type="application/zip")
        response["Content-Length"] = os.path.getsize(export_file)
        response["Content-Disposition"] = "inline; filename=%s" % export_filename

        return response


def create_export_directory(tmp_dir, theme):
    export_name = get_export_name(theme)
    export_dir = os.path.join(tmp_dir, export_name)
    os.mkdir(export_dir)
    return export_dir


def get_export_name(theme):
    if theme.version:
        return "%s-%s" % (slugify(theme.name), theme.version.replace(".", "-"))
    return slugify(theme.name)


def create_theme_manifest(theme):
    return {
        "name": theme.name,
        "version": theme.version,
        "author": theme.author,
        "url": theme.url,
        "css": [],
        "media": [],
    }


def write_theme_css(export_dir, theme):
    files_dir = create_sub_directory(export_dir, "css")
    files = []

    for css in theme.css.all():
        if css.url:
            files.append({"name": css.name, "url": css.url})
        else:
            files.append(
                {"name": css.name, "path": copy_asset_file(files_dir, css.source_file)}
            )

    return files


def write_theme_media(export_dir, theme):
    files_dir = create_sub_directory(export_dir, "media")
    files = []

    for media in theme.media.all():
        files.append(
            {
                "name": media.name,
                "type": media.type,
                "path": copy_asset_file(files_dir, media.file),
            }
        )

    return files


def create_sub_directory(export_dir, dirname):
    new_dir = os.path.join(export_dir, dirname)
    os.mkdir(new_dir)
    return new_dir


def copy_asset_file(export_dir, asset_file):
    filename = os.path.split(asset_file.name)[-1]
    dst_path = os.path.join(export_dir, filename)
    with open(dst_path, "wb") as fp:
        for chunk in asset_file.chunks():
            fp.write(chunk)
    return filename


def write_theme_manifest(export_dir, manifest):
    manifest_path = os.path.join(export_dir, "manifest.json")
    with open(manifest_path, "w") as fp:
        json.dump(manifest, fp, ensure_ascii=False, indent=2)


def zip_theme_export(tmp_dir, export_dir):
    dir_name = os.path.split(export_dir)[-1]
    return shutil.make_archive(export_dir, "zip", tmp_dir, dir_name)



from django.http import Http404

from ...acl.objectacl import add_acl_to_obj
from ...categories.models import Category
from ...categories.permissions import allow_browse_category, allow_see_category
from ...categories.serializers import CategorySerializer
from ...core.shortcuts import validate_slug
from ...core.viewmodel import ViewModel as BaseViewModel
from ..permissions import allow_use_private_threads

__all__ = ["ThreadsRootCategory", "ThreadsCategory", "PrivateThreadsCategory"]


class ViewModel(BaseViewModel):
    def __init__(self, request, **kwargs):
        self.request = request

        self._categories = self.get_categories(request)
        add_acl_to_obj(request.user_acl, self._categories)

        self._model = self.get_category(request, self._categories, **kwargs)

        self._subcategories = list(filter(self._model.has_child, self._categories))
        self._children = list(
            filter(lambda s: s.parent_id == self._model.pk, self._subcategories)
        )

    @property
    def categories(self):
        return self._categories

    @property
    def subcategories(self):
        return self._subcategories

    @property
    def children(self):
        return self._children

    def get_categories(self, request):
        raise NotImplementedError(
            "Category view model has to implement get_categories(request)"
        )

    def get_category(self, request, categories, **kwargs):
        return categories[0]

    def get_frontend_context(self):
        return {
            "CATEGORIES": BasicCategorySerializer(
                self._categories,
                context={"settings": self.request.settings},
                many=True,
            ).data
        }

    def get_template_context(self):
        top_category = None
        sub_category = None

        if self._model.level == 1:
            top_category = self._model
        elif self._model.level == 2:
            top_category = self._model.parent
            sub_category = self._model

        return {
            "category": self._model,
            "top_category": top_category,
            "sub_category": sub_category,
            "categories": self._categories,
            "subcategories": self._children,
        }


class ThreadsRootCategory(ViewModel):
    def get_categories(self, request):
        return [Category.objects.root_category()] + list(
            Category.objects.all_categories()
            .filter(id__in=request.user_acl["visible_categories"])
            .select_related("parent")
        )


class ThreadsCategory(ThreadsRootCategory):
    @property
    def level(self):
        return self._model.level

    def get_category(self, request, categories, **kwargs):
        for category in categories:
            if category.pk == int(kwargs["pk"]):
                if not category.special_role:
                    # check permissions for non-special categories
                    allow_see_category(request.user_acl, category)
                    allow_browse_category(request.user_acl, category)

                if "slug" in kwargs:
                    validate_slug(category, kwargs["slug"])

                return category
        raise Http404()


class PrivateThreadsCategory(ViewModel):
    def get_categories(self, request):
        return [Category.objects.private_threads()]

    def get_category(self, request, categories, **kwargs):
        allow_use_private_threads(request.user_acl)

        return categories[0]


BasicCategorySerializer = CategorySerializer.subset_fields(
    "id",
    "parent",
    "name",
    "short_name",
    "color",
    "description",
    "is_closed",
    "css_class",
    "level",
    "lft",
    "rght",
    "is_read",
    "url",
)



class MutableFields:
    @classmethod
    def subset_fields(cls, *fields):
        fields_in_name = [f.title().replace("_", "") for f in fields]
        name = "%s%sSubset" % (cls.__name__, "".join(fields_in_name)[:100])

        class Meta(cls.Meta):
            pass

        Meta.fields = list(fields)

        return type(name, (cls,), {"Meta": Meta})

    @classmethod
    def exclude_fields(cls, *fields):
        final_fields = []
        for field in cls.Meta.fields:
            if field not in fields:
                final_fields.append(field)

        fields_in_name = [f.title().replace("_", "") for f in final_fields]
        name = "%s%sSubset" % (cls.__name__, "".join(fields_in_name)[:100])

        class Meta(cls.Meta):
            pass

        Meta.fields = list(final_fields)

        return type(name, (cls,), {"Meta": Meta})

    @classmethod
    def extend_fields(cls, *fields):
        final_fields = list(cls.Meta.fields)
        for field in fields:
            if field not in final_fields:
                final_fields.append(field)

        fields_in_name = [f.title().replace("_", "") for f in final_fields]
        name = "%s%sSubset" % (cls.__name__, "".join(fields_in_name)[:100])

        class Meta(cls.Meta):
            pass

        Meta.fields = list(final_fields)

        return type(name, (cls,), {"Meta": Meta})



from django.urls import path

from .views import watch_private_thread, watch_thread

urlpatterns = [
    path(
        "private-threads/<int:thread_id>/watch/",
        watch_private_thread,
        name="private-thread-watch",
    ),
    path("threads/<int:thread_id>/watch/", watch_thread, name="thread-watch"),
]



from rest_framework import serializers

from ...categories.serializers import CategorySerializer
from ...core.serializers import MutableFields
from ...users.serializers import UserSerializer
from ..models import Post
from .post import PostSerializer

__all__ = ["FeedSerializer"]

FeedUserSerializer = UserSerializer.subset_fields(
    "id", "username", "avatars", "url", "title", "rank"
)

FeedCategorySerializer = CategorySerializer.subset_fields("name", "css_class", "url")


class FeedSerializer(PostSerializer, MutableFields):
    poster = FeedUserSerializer(many=False, read_only=True)
    category = FeedCategorySerializer(many=False, read_only=True)

    thread = serializers.SerializerMethodField()

    class Meta:
        model = Post
        fields = PostSerializer.Meta.fields + ["category", "thread"]

    def get_thread(self, obj):
        return {"title": obj.thread.title, "url": obj.thread.get_absolute_url()}


FeedSerializer = FeedSerializer.exclude_fields("is_liked", "is_new", "is_read")



from django.urls import path

from .views import disable_email_notifications, notification, notifications

urlpatterns = [
    path(
        "notifications/disable-email/<int:watched_thread_id>/<str:secret>/",
        disable_email_notifications,
        name="notifications-disable-email",
    ),
    path("notification/<int:notification_id>/", notification, name="notification"),
    path("notifications/", notifications, name="notifications"),
    path("notifications/unread/", notifications, name="notifications-unread"),
    path("notifications/read/", notifications, name="notifications-read"),
]



from hashlib import md5

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.translation import pgettext_lazy


def get_data_upload_to(instance, filename):
    user_id_hexdigest = md5(str(instance.user_id).encode()).hexdigest()
    return "data-downloads/%s/%s/%s.zip" % (
        user_id_hexdigest,
        get_random_string(64),
        instance.user.slug,
    )


class DataDownload(models.Model):
    STATUS_PENDING = 0
    STATUS_PROCESSING = 1
    STATUS_READY = 2
    STATUS_EXPIRED = 3

    STATUS_CHOICES = [
        (
            STATUS_PENDING,
            pgettext_lazy("user data download status", "Pending"),
        ),
        (
            STATUS_PROCESSING,
            pgettext_lazy("user data download status", "Processing"),
        ),
        (
            STATUS_READY,
            pgettext_lazy("user data download status", "Ready"),
        ),
        (
            STATUS_EXPIRED,
            pgettext_lazy("user data download status", "Expired"),
        ),
    ]

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    status = models.PositiveIntegerField(
        default=STATUS_PENDING, choices=STATUS_CHOICES, db_index=True
    )
    requester = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    requester_name = models.CharField(max_length=255)
    requested_on = models.DateTimeField(default=timezone.now, db_index=True)
    expires_on = models.DateTimeField(default=timezone.now)
    file = models.FileField(
        upload_to=get_data_upload_to, max_length=255, null=True, blank=True
    )

    class Meta:
        ordering = ["-pk"]

    def delete(self, *args, **kwargs):
        if self.file:
            self.file.delete(save=False)
        super().delete(*args, **kwargs)



from django.contrib.auth import get_user_model
from django.dispatch import receiver
from django.db.models import Q
from django.db.models.signals import pre_delete

from ..core.pgutils import chunk_queryset
from ..users.signals import (
    anonymize_user_data,
    archive_user_data,
    delete_user_content,
    username_changed,
)
from .registry import registry
from .models import Notification, WatchedThread


@receiver(archive_user_data)
def archive_user_notifications(sender, archive=None, **kwargs):
    queryset = Notification.objects.filter(Q(user=sender) | Q(actor=sender)).order_by(
        "id"
    )

    for notification in chunk_queryset(queryset):
        item_name = notification.created_at.strftime("%H%M%S-notification")
        archive.add_text(
            item_name,
            registry.get_message(notification),
            date=notification.created_at,
        )


@receiver([anonymize_user_data, username_changed])
def update_actor_name(sender, **kwargs):
    Notification.objects.filter(actor=sender).update(actor_name=sender.username)


@receiver(delete_user_content)
def delete_user_data(sender, **kwargs):
    Notification.objects.filter(Q(user=sender) | Q(actor=sender)).delete()
    WatchedThread.objects.filter(user=sender).delete()


@receiver(pre_delete, sender=get_user_model())
def delete_user_account(sender, *, instance, **kwargs):
    Notification.objects.filter(actor=instance).update(actor=None)
    Notification.objects.filter(user=instance).delete()
    WatchedThread.objects.filter(user=instance).delete()



from django.core.exceptions import PermissionDenied
from django.core.paginator import EmptyPage, InvalidPage
from django.db.models import Q
from django.http import Http404
from django.utils.translation import pgettext, pgettext_lazy

from ...acl.objectacl import add_acl_to_obj
from ...core.cursorpagination import get_page
from ...notifications.models import WatchedThread
from ...notifications.threads import get_watched_threads
from ...readtracker import threadstracker
from ...readtracker.cutoffdate import get_cutoff_date
from ..models import Post, Thread
from ..participants import make_participants_aware
from ..permissions import exclude_invisible_posts, exclude_invisible_threads
from ..serializers import ThreadsListSerializer
from ..utils import add_categories_to_items

__all__ = ["ForumThreads", "PrivateThreads", "filter_read_threads_queryset"]

LISTS_NAMES = {
    "all": None,
    "my": pgettext_lazy("threads list", "Your threads"),
    "new": pgettext_lazy("threads list", "New threads"),
    "unread": pgettext_lazy("threads list", "Unread threads"),
    "watched": pgettext_lazy("threads list", "Watched threads"),
    "unapproved": pgettext_lazy("threads list", "Unapproved content"),
}

LIST_DENIED_MESSAGES = {
    "my": pgettext_lazy(
        "threads list",
        "You have to sign in to see list of threads that you have started.",
    ),
    "new": pgettext_lazy(
        "threads list",
        "You have to sign in to see list of threads you haven't read.",
    ),
    "unread": pgettext_lazy(
        "threads list",
        "You have to sign in to see list of threads with new replies.",
    ),
    "watched": pgettext_lazy(
        "threads list",
        "You have to sign in to see list of threads you are watching.",
    ),
    "unapproved": pgettext_lazy(
        "threads list",
        "You have to sign in to see list of threads with unapproved posts.",
    ),
}


class ViewModel:
    def __init__(self, request, category, list_type, start=0):
        self.request = request

        self.allow_see_list(request, category, list_type)

        category_model = category.unwrap()

        base_queryset = self.get_base_queryset(request, category.categories, list_type)
        base_queryset = base_queryset.select_related("starter", "last_poster")

        threads_categories = [category_model] + category.subcategories

        threads_queryset = self.get_remaining_threads_queryset(
            base_queryset, category_model, threads_categories
        )

        try:
            list_page = get_page(
                threads_queryset,
                "-last_post_id",
                request.settings.threads_per_page,
                start,
            )
        except (EmptyPage, InvalidPage):
            raise Http404()

        if list_page.first:
            pinned_threads = list(
                self.get_pinned_threads(
                    base_queryset, category_model, threads_categories
                )
            )
            threads = list(pinned_threads) + list(list_page.object_list)
        else:
            threads = list(list_page.object_list)

        add_categories_to_items(category_model, category.categories, threads)
        add_acl_to_obj(request.user_acl, threads)

        if list_type == "watched" and request.user.is_authenticated:
            self.watched_threads = get_watched_threads(request.user, threads)
        else:
            self.watched_threads = {}

        if list_type in ("new", "unread"):
            # we already know all threads on list are unread
            for thread in threads:
                thread.is_read = False
                thread.is_new = True
        else:
            threadstracker.make_read_aware(request, threads)

        self.filter_threads(request, threads)

        # set state on object for easy access from hooks
        self.category = category
        self.threads = threads
        self.list_type = list_type
        self.list_page = list_page

    def allow_see_list(self, request, category, list_type):
        if list_type not in LISTS_NAMES:
            raise Http404()

        if request.user.is_anonymous:
            if list_type in LIST_DENIED_MESSAGES:
                raise PermissionDenied(LIST_DENIED_MESSAGES[list_type])
        else:
            has_permission = request.user_acl["can_see_unapproved_content_lists"]
            if list_type == "unapproved" and not has_permission:
                raise PermissionDenied(
                    pgettext(
                        "threads list",
                        "You don't have permission to see unapproved content lists.",
                    )
                )

    def get_list_name(self, list_type):
        return LISTS_NAMES[list_type]

    def get_base_queryset(self, request, threads_categories, list_type):
        return get_threads_queryset(request, threads_categories, list_type).order_by(
            "-last_post_id"
        )

    def get_pinned_threads(self, queryset, category, threads_categories):
        return []

    def get_remaining_threads_queryset(self, queryset, category, threads_categories):
        return []

    def filter_threads(self, request, threads):
        pass  # hook for custom thread types to add features to extend threads

    def get_frontend_context(self):
        return {
            "THREADS": {
                "results": ThreadsListSerializer(
                    self.threads,
                    many=True,
                    context={
                        "settings": self.request.settings,
                        "watched_threads": self.watched_threads,
                    },
                ).data,
                "subcategories": [c.pk for c in self.category.children],
                "next": self.list_page.next,
            }
        }

    def get_template_context(self):
        return {
            "list_name": self.get_list_name(self.list_type),
            "list_type": self.list_type,
            "list_page": self.list_page,
            "threads": self.threads,
        }


class ForumThreads(ViewModel):
    def get_pinned_threads(self, queryset, category, threads_categories):
        if category.level:
            return list(queryset.filter(weight=2)) + list(
                queryset.filter(weight=1, category__in=threads_categories)
            )
        return queryset.filter(weight=2)

    def get_remaining_threads_queryset(self, queryset, category, threads_categories):
        if category.level:
            return queryset.filter(weight=0, category__in=threads_categories)
        return queryset.filter(weight__lt=2, category__in=threads_categories)


class PrivateThreads(ViewModel):
    def get_base_queryset(self, request, threads_categories, list_type):
        queryset = super().get_base_queryset(request, threads_categories, list_type)

        # limit queryset to threads we are participant of
        participated_threads = request.user.threadparticipant_set.values("thread_id")

        if request.user_acl["can_moderate_private_threads"]:
            queryset = queryset.filter(
                Q(id__in=participated_threads) | Q(has_reported_posts=True)
            )
        else:
            queryset = queryset.filter(id__in=participated_threads)

        return queryset

    def get_remaining_threads_queryset(self, queryset, category, threads_categories):
        return queryset.filter(category__in=threads_categories)

    def filter_threads(self, request, threads):
        make_participants_aware(request.user, threads)


def get_threads_queryset(request, categories, list_type):
    queryset = exclude_invisible_threads(request.user_acl, categories, Thread.objects)
    if list_type == "all":
        return queryset
    return filter_threads_queryset(request, categories, list_type, queryset)


def filter_threads_queryset(request, categories, list_type, queryset):
    if list_type == "my":
        return queryset.filter(starter=request.user)
    if list_type == "watched":
        watched_threads = WatchedThread.objects.filter(user=request.user).values(
            "thread_id"
        )
        return queryset.filter(id__in=watched_threads)
    if list_type == "unapproved":
        return queryset.filter(has_unapproved_posts=True)
    if list_type in ("new", "unread"):
        return filter_read_threads_queryset(request, categories, list_type, queryset)
    return queryset


def filter_read_threads_queryset(request, categories, list_type, queryset):
    # grab cutoffs for categories
    cutoff_date = get_cutoff_date(request.settings, request.user)

    visible_posts = Post.objects.filter(posted_on__gt=cutoff_date)
    visible_posts = exclude_invisible_posts(request.user_acl, categories, visible_posts)

    queryset = queryset.filter(id__in=visible_posts.distinct().values("thread"))

    read_posts = visible_posts.filter(id__in=request.user.postread_set.values("post"))

    if list_type == "new":
        # new threads have no entry in reads table
        return queryset.exclude(id__in=read_posts.distinct().values("thread"))

    if list_type == "unread":
        # unread threads were read in past but have new posts
        unread_posts = visible_posts.exclude(
            id__in=request.user.postread_set.values("post")
        )
        queryset = queryset.filter(id__in=read_posts.distinct().values("thread"))
        queryset = queryset.filter(id__in=unread_posts.distinct().values("thread"))
        return queryset



from django import template
from django.utils.translation import npgettext, pgettext

register = template.Library()


@register.simple_tag
def likes_label(post):
    last_likes = post.last_likes or []

    usernames = []
    for like in last_likes[:3]:
        usernames.append(like["username"])

    if len(usernames) == 1:
        return pgettext("post likes message", "%(user)s likes this.") % {
            "user": usernames[0]
        }

    hidden_likes = post.likes - len(usernames)
    if len(last_likes) < 4:
        usernames_string = humanize_usernames_list(usernames)
    else:
        usernames_string = ", ".join(usernames)

    if not hidden_likes:
        return pgettext("post likes message", "%(users)s like this.") % {
            "users": usernames_string
        }

    label = npgettext(
        "post likes message",
        "%(users)s and %(likes)s other user like this.",
        "%(users)s and %(likes)s other users like this.",
        hidden_likes,
    )
    formats = {"users": usernames_string, "likes": hidden_likes}

    return label % formats


def humanize_usernames_list(usernames):
    formats = {"users": ", ".join(usernames[:-1]), "last_user": usernames[-1]}

    return pgettext("post likes message", "%(users)s and %(last_user)s") % formats



from django.urls import path
from django.contrib import admin as djadmin
from django.contrib.auth import get_user_model
from django.utils.translation import pgettext_lazy

from .djangoadmin import UserAdminModel
from .views.bans import BansList, DeleteBan, EditBan, NewBan
from .views.datadownloads import DataDownloadsList, RequestDataDownloads
from .views.ranks import (
    DefaultRank,
    DeleteRank,
    EditRank,
    MoveDownRank,
    MoveUpRank,
    NewRank,
    RanksList,
    RankUsers,
)
from .views.users import EditUser, NewUser, UsersList

djadmin.site.register(model_or_iterable=get_user_model(), admin_class=UserAdminModel)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Users section
        urlpatterns.namespace("users/", "users")

        # Accounts
        urlpatterns.patterns(
            "users",
            path("", UsersList.as_view(), name="index"),
            path("<int:page>/", UsersList.as_view(), name="index"),
            path("new/", NewUser.as_view(), name="new"),
            path("edit/<int:pk>/", EditUser.as_view(), name="edit"),
        )

        # Bans
        urlpatterns.namespace("bans/", "bans", "users")
        urlpatterns.patterns(
            "users:bans",
            path("", BansList.as_view(), name="index"),
            path("<int:page>/", BansList.as_view(), name="index"),
            path("new/", NewBan.as_view(), name="new"),
            path("edit/<int:pk>/", EditBan.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteBan.as_view(), name="delete"),
        )

        # Data Downloads
        urlpatterns.namespace("data-downloads/", "data-downloads", "users")
        urlpatterns.patterns(
            "users:data-downloads",
            path("", DataDownloadsList.as_view(), name="index"),
            path("<int:page>/", DataDownloadsList.as_view(), name="index"),
            path("request/", RequestDataDownloads.as_view(), name="request"),
        )

        # Ranks
        urlpatterns.namespace("ranks/", "ranks")
        urlpatterns.patterns(
            "ranks",
            path("", RanksList.as_view(), name="index"),
            path("new/", NewRank.as_view(), name="new"),
            path("edit/<int:pk>/", EditRank.as_view(), name="edit"),
            path("default/<int:pk>/", DefaultRank.as_view(), name="default"),
            path("move/down/<int:pk>/", MoveDownRank.as_view(), name="down"),
            path("move/up/<int:pk>/", MoveUpRank.as_view(), name="up"),
            path("users/<int:pk>/", RankUsers.as_view(), name="users"),
            path("delete/<int:pk>/", DeleteRank.as_view(), name="delete"),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=pgettext_lazy("admin node", "Users"),
            icon="fa fa-users",
            after="index",
            namespace="users",
        )

        site.add_node(
            name=pgettext_lazy("admin node", "Bans"),
            parent="users",
            namespace="bans",
        )

        site.add_node(
            name=pgettext_lazy("admin node", "Data downloads"),
            parent="users",
            after="bans:index",
            namespace="data-downloads",
        )

        site.add_node(
            name=pgettext_lazy("admin node", "Ranks"),
            icon="fas fa-shield-alt",
            after="users:index",
            namespace="ranks",
        )



from datetime import timedelta

from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ...threads.test import reply_thread
from ..activepostersranking import (
    build_active_posters_ranking,
    get_active_posters_ranking,
)
from ..models import ActivityRanking
from ..test import create_test_user


def test_ranking_is_emptied_if_no_users_exist(post):
    assert not build_active_posters_ranking()


def test_ranking_is_emptied_if_no_posts_exist(user):
    assert not build_active_posters_ranking()


@override_dynamic_settings(top_posters_ranking_length=5)
def test_recent_post_by_user_counts_to_ranking(thread, user):
    reply_thread(thread, poster=user)
    assert build_active_posters_ranking()


@override_dynamic_settings(top_posters_ranking_length=5)
def test_recent_post_by_removed_user_doesnt_count_to_ranking(thread):
    reply_thread(thread)
    assert not build_active_posters_ranking()


@override_dynamic_settings(top_posters_ranking_length=5)
def test_old_post_by_user_doesnt_count_to_ranking(thread, user):
    reply_thread(thread, poster=user, posted_on=timezone.now() - timedelta(days=6))
    assert not build_active_posters_ranking()


@override_dynamic_settings(top_posters_ranking_length=5)
def test_event_posts_doesnt_count_to_ranking(thread, user):
    reply_thread(thread, poster=user, is_event=True)
    assert not build_active_posters_ranking()


@override_dynamic_settings(top_posters_ranking_size=2)
def test_ranking_size_is_limited(thread):
    for i in range(3):
        user = create_test_user("User%s" % i, "user%s@example.com" % i)
        reply_thread(thread, poster=user)
    assert len(build_active_posters_ranking()) == 2


@override_dynamic_settings(top_posters_ranking_size=2)
def test_old_ranking_is_removed_during_build(user):
    ActivityRanking.objects.create(user=user, score=1)
    build_active_posters_ranking()
    assert not ActivityRanking.objects.exists()


def test_empty_ranking_is_returned_from_db(db):
    assert get_active_posters_ranking() == {"users": [], "users_count": 0}


def test_ranking_is_returned_from_db(user):
    ActivityRanking.objects.create(user=user, score=1)
    assert get_active_posters_ranking() == {"users": [user], "users_count": 1}


def test_ranked_user_is_annotated_with_score(user):
    ActivityRanking.objects.create(user=user, score=1)
    ranked_user = get_active_posters_ranking()["users"][0]
    assert ranked_user.score == 1


def test_ranked_users_are_ordered_by_score(user, other_user):
    ActivityRanking.objects.create(user=user, score=1)
    ActivityRanking.objects.create(user=other_user, score=2)
    assert get_active_posters_ranking() == {
        "users": [other_user, user],
        "users_count": 2,
    }



from ..uploadto import add_hash_to_filename


def test_hash_is_added_before_file_extension():
    filename = add_hash_to_filename("hash", "test.jpg")
    assert filename == "test.hash.jpg"


def test_hash_is_added_before_file_extension_in_filename_with_multiple_dots():
    filename = add_hash_to_filename("hash", "test.image.jpg")
    assert filename == "test.image.hash.jpg"


def test_hash_is_not_added_to_filename_if_it_already_contains_it():
    filename = add_hash_to_filename("hash", "test.hash.jpg")
    assert filename == "test.hash.jpg"



from django.urls import path
from django.utils.translation import pgettext_lazy

from .views import icons_admin


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Icons
        urlpatterns.namespace("icons/", "icons", "settings")
        urlpatterns.patterns("settings:icons", path("", icons_admin, name="index"))

    def register_navigation_nodes(self, site):
        site.add_node(
            name=pgettext_lazy("admin node", "Icons"),
            description=pgettext_lazy(
                "admin node", "Upload favicon and application icon for the site."
            ),
            parent="settings",
            namespace="icons",
        )



from datetime import timedelta
from typing import TYPE_CHECKING, Dict, Iterable, Optional

from django.db.models import IntegerChoices
from django.utils.translation import pgettext, pgettext_lazy

from ..acl.useracl import get_user_acl
from ..categories.enums import CategoryTree
from ..conf.dynamicsettings import DynamicSettings
from ..core.mail import build_mail
from ..threads.models import Post, Thread, ThreadParticipant
from ..threads.permissions.privatethreads import (
    can_see_private_thread,
    can_use_private_threads,
)
from ..threads.permissions.threads import (
    can_see_post,
    can_see_thread,
    exclude_invisible_posts,
)
from .verbs import NotificationVerb
from .models import Notification, WatchedThread
from .users import notify_user

if TYPE_CHECKING:
    from ..users.models import User


class ThreadNotifications(IntegerChoices):
    NONE = 0, pgettext_lazy("notification type", "Don't notify")
    SITE_ONLY = 1, pgettext_lazy("notification type", "Notify on site only")
    SITE_AND_EMAIL = 2, pgettext_lazy(
        "notification type", "Notify on site and with e-mail"
    )


def get_watched_thread(user: "User", thread: Thread) -> Optional[WatchedThread]:
    """Returns watched thread entry for given user and thread combo.

    If multiple entries are returned, it quietly heals this user's watching entry.
    """
    watched_threads = WatchedThread.objects.filter(user=user, thread=thread).order_by(
        "id"
    )[:2]
    if not watched_threads:
        return None

    watched_thread = watched_threads[0]
    if len(watched_threads) > 1:
        WatchedThread.objects.filter(user=user, thread=thread).exclude(
            id=watched_thread.id
        ).delete()

    return watched_thread


def get_watched_threads(
    user: "User", threads: Iterable[Thread]
) -> Dict[int, ThreadNotifications]:
    queryset = (
        WatchedThread.objects.filter(user=user, thread__in=threads)
        .order_by("-id")
        .values_list("thread_id", "send_emails")
    )

    return {
        thread_id: ThreadNotifications.SITE_AND_EMAIL
        if send_emails
        else ThreadNotifications.SITE_ONLY
        for thread_id, send_emails in queryset
    }


def watch_started_thread(user: "User", thread: Thread):
    if user.watch_started_threads:
        WatchedThread.objects.create(
            user=user,
            category=thread.category,
            thread=thread,
            send_emails=user.watch_started_threads
            == ThreadNotifications.SITE_AND_EMAIL,
        )


def watch_replied_thread(user: "User", thread: Thread) -> WatchedThread | None:
    if not user.watch_replied_threads:
        return

    if WatchedThread.objects.filter(user=user, thread=thread).exists():
        return None

    send_emails = user.watch_replied_threads == ThreadNotifications.SITE_AND_EMAIL

    return WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        send_emails=send_emails,
    )


def notify_watcher_on_new_thread_reply(
    watched_thread: WatchedThread,
    post: Post,
    cache_versions: Dict[str, str],
    settings: DynamicSettings,
):
    is_private = post.category.tree_id == CategoryTree.PRIVATE_THREADS
    user_acl = get_user_acl(watched_thread.user, cache_versions)

    if not user_can_see_post(watched_thread.user, user_acl, post, is_private):
        return  # Skip this watcher because they can't see the post

    if user_has_other_unread_posts(watched_thread, user_acl, post, is_private):
        return  # We only notify on first unread post

    notify_user(
        watched_thread.user,
        NotificationVerb.REPLIED,
        post.poster,
        post.category,
        post.thread,
        post,
    )

    if watched_thread.send_emails:
        email_watcher_on_new_thread_reply(watched_thread, post, settings)


def email_watcher_on_new_thread_reply(
    watched_thread: WatchedThread,
    post: Post,
    settings: DynamicSettings,
):
    subject = pgettext(
        "new thread reply email subject", "%(thread)s - new reply by %(user)s"
    ) % {
        "user": post.poster.username,
        "thread": post.thread.title,
    }

    message = build_mail(
        watched_thread.user,
        subject,
        "misago/emails/thread/reply",
        sender=post.poster,
        context={
            "settings": settings,
            "watched_thread": watched_thread,
            "thread": post.thread,
            "post": post,
        },
    )

    message.send()


def user_has_other_unread_posts(
    watched_thread: WatchedThread,
    user_acl: dict,
    post: Post,
    is_private: bool,
) -> bool:
    posts_queryset = Post.objects.filter(
        id__lt=post.id,
        thread=post.thread,
        posted_on__gt=watched_thread.read_at,
    ).exclude(poster=watched_thread.user)

    if not is_private:
        posts_queryset = exclude_invisible_posts(
            user_acl, post.category, posts_queryset
        )

    return posts_queryset.exists()


def user_can_see_post(
    user: "User",
    user_acl: dict,
    post: Post,
    is_private: bool,
) -> bool:
    if is_private:
        return user_can_see_post_in_private_thread(user, user_acl, post)

    return can_see_thread(user_acl, post.thread) and can_see_post(user_acl, post)


def user_can_see_post_in_private_thread(
    user: "User", user_acl: dict, post: Post
) -> bool:
    if not can_use_private_threads(user_acl):
        return False

    is_participant = ThreadParticipant.objects.filter(
        thread=post.thread, user=user
    ).exists()
    if not can_see_private_thread(user_acl, post.thread, is_participant):
        return False

    return True


def notify_participant_on_new_private_thread(
    user: "User",
    actor: "User",
    thread: Thread,
    cache_versions: dict[str, str],
    settings: DynamicSettings,
):
    user_acl = get_user_acl(user, cache_versions)

    if not user_can_see_private_thread(user, user_acl, thread):
        return False  # User can't see private thread

    actor_is_followed = user.is_following(actor)

    watched_thread = watch_new_private_thread(user, thread, actor_is_followed)

    if actor_is_followed:
        notification = user.notify_new_private_threads_by_followed
    else:
        notification = user.notify_new_private_threads_by_other_users

    if notification == ThreadNotifications.NONE:
        return  # User has disabled notifications for new private threads

    has_unread_notification = Notification.objects.filter(
        user=user,
        verb=NotificationVerb.INVITED,
        post=thread.first_post,
        is_read=False,
    ).exists()

    if has_unread_notification:
        return  # Don't spam users with notifications

    notify_user(
        user,
        NotificationVerb.INVITED,
        actor,
        thread.category,
        thread,
        thread.first_post,
    )

    if notification == ThreadNotifications.SITE_AND_EMAIL:
        email_participant_on_new_private_thread(user, actor, watched_thread, settings)


def email_participant_on_new_private_thread(
    user: "User",
    actor: "User",
    watched_thread: WatchedThread,
    settings: DynamicSettings,
):
    thread = watched_thread.thread

    subject = pgettext(
        "new private thread email subject",
        '%(user)s has invited you to participate in private thread "%(thread)s"',
    )
    subject_formats = {"thread": thread.title, "user": actor.username}

    message = build_mail(
        user,
        subject % subject_formats,
        "misago/emails/privatethread/added",
        sender=actor,
        context={
            "settings": settings,
            "watched_thread": watched_thread,
            "thread": thread,
        },
    )

    message.send()


def watch_new_private_thread(
    user: "User", thread: Thread, from_followed: bool
) -> WatchedThread | None:
    if from_followed:
        notifications = user.watch_new_private_threads_by_followed
    else:
        notifications = user.watch_new_private_threads_by_other_users

    if not notifications:
        return None

    send_emails = notifications == ThreadNotifications.SITE_AND_EMAIL

    # Set thread read date in past to prevent first reply to thread
    # From triggering extra notification
    read_at = thread.started_on - timedelta(seconds=5)

    if watched_thread := get_watched_thread(user, thread):
        watched_thread.read_at = read_at
        watched_thread.save(update_fields=["read_at"])
        return watched_thread

    return WatchedThread.objects.create(
        user=user,
        category=thread.category,
        thread=thread,
        send_emails=send_emails,
        read_at=read_at,
    )


def user_can_see_private_thread(user: "User", user_acl: dict, thread: Thread) -> bool:
    if not can_use_private_threads(user_acl):
        return False

    is_participant = ThreadParticipant.objects.filter(thread=thread, user=user).exists()
    if not can_see_private_thread(user_acl, thread, is_participant):
        return False

    return True



# Generated by Django 3.2.15 on 2023-04-02 17:57

from django.db import migrations

from ...notifications.threads import ThreadNotifications
from ..hydrators import dehydrate_value


settings = [
    {
        "setting": "watch_started_threads",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "watch_replied_threads",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "watch_new_private_threads_by_followed",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "watch_new_private_threads_by_other_users",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "notify_new_private_threads_by_followed",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "notify_new_private_threads_by_other_users",
        "python_type": "int",
        "dry_value": ThreadNotifications.SITE_AND_EMAIL,
        "is_public": False,
    },
    {
        "setting": "delete_notifications_older_than",
        "python_type": "int",
        "dry_value": 60,
    },
]


def create_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    for setting in settings:
        data = setting.copy()
        if "python_type" in data and "dry_value" in data:
            data["dry_value"] = dehydrate_value(data["python_type"], data["dry_value"])

        Setting.objects.create(**setting)


def delete_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    Setting.objects.filter(
        setting__in=[setting["setting"] for setting in settings]
    ).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("misago_conf", "0010_add_admin_link_setting"),
    ]

    operations = [
        migrations.RunPython(create_settings, delete_settings),
    ]



from django.urls import include, path
from django.views.generic import TemplateView

from . import hooks
from .conf import settings
from .core.views import forum_index

app_name = "misago"

# Register Misago Apps
urlpatterns = [
    path("", include("misago.analytics.urls")),
    path("", include("misago.legal.urls")),
    path("", include("misago.users.urls")),
    path("", include("misago.categories.urls")),
    path("", include("misago.threads.urls")),
    path("", include("misago.notifications.urls")),
    path("", include("misago.search.urls")),
    path("", include("misago.oauth2.urls")),
    path("", include("misago.socialauth.urls")),
    path("", include("misago.healthcheck.urls")),
    # default robots.txt
    path(
        "robots.txt",
        TemplateView.as_view(
            content_type="text/plain", template_name="misago/robots.txt"
        ),
    ),
    # "misago:index" link symbolises "root" of Misago links space
    # any request with path that falls below this one is assumed to be directed
    # at Misago and will be handled by misago.views.exceptionhandler if it
    # results in Http404 or PermissionDenied exception
    path("", forum_index, name="index"),
]


# Register API
apipatterns = hooks.apipatterns + [
    path("", include("misago.legal.urls.api")),
    path("", include("misago.markup.urls")),
    path("", include("misago.threads.urls.api")),
    path("", include("misago.users.urls.api")),
    path("", include("misago.search.urls.api")),
]

urlpatterns += [
    path("api/", include((apipatterns, "api"), namespace="api")),
    path("api/v2/", include("misago.apiv2.urls")),
]


# Register Misago ACP
if settings.MISAGO_ADMIN_PATH:
    # Admin patterns recognised by Misago
    adminpatterns = [path("", include("misago.admin.urls"))]

    admin_prefix = f"{settings.MISAGO_ADMIN_PATH}/"
    urlpatterns += [
        path(admin_prefix, include((adminpatterns, "admin"), namespace="admin"))
    ]



class Page:
    """
    Misago page utility

    Allows for adding custom views to "sectioned" pages like
    User Control Panel, Users List or Threads Lists
    """

    def __init__(self, name):
        self._finalized = False
        self.name = name
        self._unsorted_list = []
        self._sorted_list = []

    def assert_is_finalized(self):
        if not self._finalized:
            self._finalized = True
            self._finalize()

    def _finalize(self):
        iterations = 0
        while self._unsorted_list:
            iterations += 1
            if iterations > 512:
                message = (
                    "%s page hierarchy is invalid or too complex to resolve. "
                    "Sections left: %s"
                )
                raise ValueError(message % self._unsorted_list)

            for index, section in enumerate(self._unsorted_list):
                if section["after"]:
                    section_added = self._insert_section(
                        section, after=section["after"]
                    )
                elif section["before"]:
                    section_added = self._insert_section(
                        section, before=section["before"]
                    )
                else:
                    section_added = self._insert_section(section)

                if section_added:
                    del self._unsorted_list[index]
                    break

    def _insert_section(self, inserted_section, after=None, before=None):
        if after:
            new_sorted_list = []
            for section in self._sorted_list:
                new_sorted_list.append(section)
                if section["link"] == after:
                    new_sorted_list.append(inserted_section)
                    self._sorted_list = new_sorted_list
                    return True
            return False

        if before:
            new_sorted_list = []
            for section in self._sorted_list:
                if section["link"] == before:
                    new_sorted_list.append(inserted_section)
                    new_sorted_list.append(section)
                    self._sorted_list = new_sorted_list
                    return True
                new_sorted_list.append(section)
            return False

        self._sorted_list.append(inserted_section)
        return True

    def add_section(
        self,
        link,
        after=None,
        before=None,
        visible_if=None,
        get_metadata=None,
        **kwargs
    ):
        if self._finalized:
            message = (
                "%s page was initialized already and no longer accepts new sections"
            )
            raise RuntimeError(message % self.name)

        if after and before:
            raise ValueError("after and before arguments are exclusive")

        kwargs.update(
            {
                "link": link,
                "after": after,
                "before": before,
                "visible_if": visible_if,
                "get_metadata": get_metadata,
            }
        )

        self._unsorted_list.append(kwargs)

    def _active_link_name(self, request) -> str | None:
        if not request.resolver_match:
            return None

        namespace = request.resolver_match.namespace
        url_name = request.resolver_match.url_name

        if namespace:
            active_link = "%s:%s" % (namespace, url_name)
        else:
            active_link = url_name
        return active_link

    def get_sections(self, request, *args):
        self.assert_is_finalized()
        active_link = self._active_link_name(request)
        visible_sections = []

        for section_definition in self._sorted_list:
            section = section_definition.copy()

            is_visible = True
            if section["visible_if"]:
                is_visible = section["visible_if"](request, *args)

            if is_visible:
                if section["get_metadata"]:
                    section["metadata"] = section["get_metadata"](request, *args)
                if active_link:
                    section["is_active"] = active_link.startswith(section["link"])
                else:
                    section["is_active"] = None
                visible_sections.append(section)
        return visible_sections

    def get_default_link(self):
        self.assert_is_finalized()
        return self._sorted_list[0]["link"]
